from __future__ import annotations

from http import HTTPStatus
from typing import TYPE_CHECKING

from graphql import GraphQLError

from undine.exceptions import GraphQLErrorGroup, GraphQLRequestDecodingError
from undine.http.utils import (
    graphql_error_group_response,
    graphql_error_response,
    graphql_result_response,
    parse_json_body,
    require_persisted_documents_request,
)
from undine.settings import undine_settings
from undine.utils.graphql.utils import build_response

from .utils import parse_document_map, register_persisted_documents

if TYPE_CHECKING:
    from undine.typing import DjangoRequestProtocol, DjangoResponseProtocol

__all__ = [
    "persisted_documents_view",
]


@require_persisted_documents_request
def persisted_documents_view(request: DjangoRequestProtocol) -> DjangoResponseProtocol:
    """
    View for registering persisted documents.
    Users should add permission checks.
    """
    try:
        json_data = parse_json_body(request.body)
    except GraphQLRequestDecodingError as error:
        return graphql_error_response(error, status=HTTPStatus.BAD_REQUEST)

    try:
        document_map = parse_document_map(json_data)
    except GraphQLErrorGroup as error:
        return graphql_error_group_response(error, status=HTTPStatus.BAD_REQUEST)

    if undine_settings.PERSISTED_DOCUMENTS_PERMISSION_CALLBACK is not None:
        try:
            undine_settings.PERSISTED_DOCUMENTS_PERMISSION_CALLBACK(request, document_map)
        except GraphQLError as error:
            return graphql_error_response(error, status=HTTPStatus.FORBIDDEN)
        except GraphQLErrorGroup as error:
            return graphql_error_group_response(error, status=HTTPStatus.FORBIDDEN)

    try:
        document_id_map = register_persisted_documents(document_map)
    except GraphQLErrorGroup as error:
        return graphql_error_group_response(error, status=HTTPStatus.BAD_REQUEST)

    result = build_response(data={"documents": document_id_map})
    return graphql_result_response(result)



from __future__ import annotations

import copy
from asyncio import ensure_future
from collections.abc import AsyncGenerator, AsyncIterable, AsyncIterator
from contextlib import aclosing, nullcontext
from functools import wraps
from http import HTTPStatus
from inspect import isawaitable
from typing import TYPE_CHECKING, Any

from django.core.exceptions import ValidationError
from graphql import ExecutionContext, ExecutionResult, GraphQLError, is_non_null_type, parse, validate
from graphql.execution.subscribe import execute_subscription

from undine.exceptions import (
    GraphQLAsyncNotSupportedError,
    GraphQLErrorGroup,
    GraphQLNoExecutionResultError,
    GraphQLUnexpectedError,
    GraphQLUseWebSocketsForSubscriptionsError,
)
from undine.hooks import LifecycleHookContext, LifecycleHookManager, use_lifecycle_hooks_async, use_lifecycle_hooks_sync
from undine.settings import undine_settings
from undine.utils.graphql.utils import build_response, is_subscription_operation, validate_get_request_operation
from undine.utils.graphql.validation_rules import get_validation_rules
from undine.utils.model_utils import get_validation_error_messages
from undine.utils.reflection import get_traceback

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable

    from graphql import DocumentNode, GraphQLOutputType
    from graphql.pyutils import AwaitableOrValue

    from undine.dataclasses import GraphQLHttpParams
    from undine.typing import DjangoRequestProtocol, ExecutionResultGen, P, WebSocketResult

__all__ = [
    "execute_graphql_http_async",
    "execute_graphql_http_sync",
    "execute_graphql_websocket",
]


class UndineExecutionContext(ExecutionContext):
    """Custom GraphQL execution context class."""

    def handle_field_error(self, error: GraphQLError, return_type: GraphQLOutputType) -> None:
        if isinstance(error.original_error, ValidationError):
            self.handle_django_validation_error(error, error.original_error)

        if not isinstance(error.original_error, GraphQLErrorGroup):
            return super().handle_field_error(error, return_type)

        error.original_error.located_by(error)

        if is_non_null_type(return_type):
            raise error.original_error

        for err in error.original_error.flatten():
            self.handle_field_error(err, return_type)

        return None

    def handle_django_validation_error(self, graphql_error: GraphQLError, original_error: ValidationError) -> None:
        graphql_error.extensions = graphql_error.extensions or {}
        graphql_error.extensions["status_code"] = HTTPStatus.BAD_REQUEST

        code = getattr(original_error, "code", None)
        if code:
            graphql_error.extensions["error_code"] = code.upper()

        error_messages = get_validation_error_messages(original_error)

        errors: list[GraphQLError] = []
        for field, messages in error_messages.items():
            for message in messages:
                path: list[Any] | None = graphql_error.path
                if field and graphql_error.path:
                    path = graphql_error.path + field.split(".")

                new_error = copy.deepcopy(graphql_error)
                new_error.message = message
                new_error.path = path
                errors.append(new_error)

        graphql_error.original_error = GraphQLErrorGroup(errors=errors)

    @staticmethod
    def build_response(data: dict[str, Any] | None, errors: list[GraphQLError]) -> ExecutionResult:
        for error in errors:
            extensions: dict[str, Any] = error.extensions  # type: ignore[union-attr]

            if error.original_error is None or isinstance(error.original_error, GraphQLError):
                extensions.setdefault("status_code", HTTPStatus.BAD_REQUEST)
            else:
                extensions.setdefault("status_code", HTTPStatus.INTERNAL_SERVER_ERROR)

            if undine_settings.INCLUDE_ERROR_TRACEBACK:
                extensions["traceback"] = get_traceback(error.__traceback__)

        return ExecutionContext.build_response(data, errors)


# HTTP sync execution


def raised_exceptions_as_execution_results_sync(
    func: Callable[P, ExecutionResult],
) -> Callable[P, ExecutionResult]:
    """Wraps raised exceptions as GraphQL ExecutionResults if they happen in `execute_graphql_sync`."""

    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> ExecutionResult:
        try:
            return func(*args, **kwargs)

        except GraphQLError as error:
            return build_response(errors=[error])

        except GraphQLErrorGroup as error:
            return build_response(errors=list(error.flatten()))

        except Exception as error:  # noqa: BLE001
            err = GraphQLUnexpectedError(message=str(error))
            return build_response(errors=[err])

    return wrapper


@raised_exceptions_as_execution_results_sync
def execute_graphql_http_sync(params: GraphQLHttpParams, request: DjangoRequestProtocol) -> ExecutionResult:
    """
    Executes a GraphQL operation received from an HTTP request synchronously.
    Assumes that the schema has been validated (e.g. created using `undine.schema.create_schema`).

    :param params: GraphQL request parameters.
    :param request: The Django request object to use as the GraphQL execution context value.
    """
    context = LifecycleHookContext.from_graphql_params(params=params, request=request)
    return _run_operation_sync(context)


@use_lifecycle_hooks_sync(hooks=undine_settings.OPERATION_HOOKS)
def _run_operation_sync(context: LifecycleHookContext) -> ExecutionResult:
    _parse_source_sync(context)
    if context.result is None:
        _validate_document_sync(context)
        if context.result is None:
            return _execute_sync(context)

    if context.result is None:  # pragma: no cover
        raise GraphQLNoExecutionResultError

    if isawaitable(context.result):
        ensure_future(context.result).cancel()
        raise GraphQLAsyncNotSupportedError

    return context.result


@use_lifecycle_hooks_sync(hooks=undine_settings.PARSE_HOOKS)
def _parse_source_sync(context: LifecycleHookContext) -> None:
    if context.result is not None:
        return

    if context.document is not None:
        return

    try:
        context.document = parse(
            source=context.source,
            no_location=undine_settings.NO_ERROR_LOCATION,
            max_tokens=undine_settings.MAX_TOKENS,
        )
    except GraphQLError as error:
        context.result = build_response(errors=[error])


@use_lifecycle_hooks_sync(hooks=undine_settings.VALIDATION_HOOKS)
def _validate_document_sync(context: LifecycleHookContext) -> None:
    if context.result is not None:
        return

    _validate_http(context)
    if context.result is not None:
        return

    validation_errors = validate(
        schema=undine_settings.SCHEMA,
        document_ast=context.document,  # type: ignore[arg-type]
        rules=get_validation_rules(),
        max_errors=undine_settings.MAX_ERRORS,
    )
    if validation_errors:
        context.result = build_response(errors=validation_errors)
        return


def _validate_http(context: LifecycleHookContext) -> None:
    if context.request.method == "GET":
        try:
            validate_get_request_operation(
                document=context.document,  # type: ignore[arg-type]
                operation_name=context.operation_name,
            )
        except GraphQLError as err:
            context.result = build_response(errors=[err])
            return

    if is_subscription_operation(context.document):  # type: ignore[arg-type]
        error: GraphQLError = GraphQLUseWebSocketsForSubscriptionsError()
        context.result = build_response(errors=[error])
        return


@use_lifecycle_hooks_sync(hooks=undine_settings.EXECUTION_HOOKS)
def _execute_sync(context: LifecycleHookContext) -> ExecutionResult:
    exec_context = _get_execution_context(
        document=context.document,  # type: ignore[arg-type]
        root_value=undine_settings.ROOT_VALUE,
        context_value=context.request,
        variable_values=context.variables,
        operation_name=context.operation_name,
    )
    result = _execute(exec_context)

    if result is None:  # pragma: no cover
        raise GraphQLNoExecutionResultError

    if isawaitable(result):
        ensure_future(result).cancel()
        raise GraphQLAsyncNotSupportedError

    context.result = result
    return context.result


# HTTP async execution


def raised_exceptions_as_execution_results_async(
    func: Callable[P, Awaitable[ExecutionResult]],
) -> Callable[P, Awaitable[ExecutionResult]]:
    """Wraps raised exceptions as GraphQL ExecutionResults if they happen in `execute_graphql_async`."""

    @wraps(func)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> ExecutionResult:
        try:
            return await func(*args, **kwargs)

        except GraphQLError as error:
            return build_response(errors=[error])

        except GraphQLErrorGroup as error:
            return build_response(errors=list(error.flatten()))

        except Exception as error:  # noqa: BLE001
            err = GraphQLUnexpectedError(message=str(error))
            return build_response(errors=[err])

    return wrapper


@raised_exceptions_as_execution_results_async
async def execute_graphql_http_async(params: GraphQLHttpParams, request: DjangoRequestProtocol) -> ExecutionResult:
    """
    Executes a GraphQL operation received from an HTTP request asynchronously.
    Assumes that the schema has been validated (e.g. created using `undine.schema.create_schema`).

    :param params: GraphQL request parameters.
    :param request: The Django request object to use as the GraphQL execution context value.
    """
    context = LifecycleHookContext.from_graphql_params(params=params, request=request)
    return await _run_operation_async(context)


@use_lifecycle_hooks_async(hooks=undine_settings.OPERATION_HOOKS)
async def _run_operation_async(context: LifecycleHookContext) -> ExecutionResult:
    await _parse_source_async(context)
    if context.result is None:
        await _validate_document_async(context)
        if context.result is None:
            return await _execute_async(context)

    if context.result is None:  # pragma: no cover
        raise GraphQLNoExecutionResultError

    if isinstance(context.result, AsyncIterator):
        raise GraphQLUseWebSocketsForSubscriptionsError

    if isawaitable(context.result):
        context.result = await context.result  # type: ignore[assignment]

    return context.result  # type: ignore[return-value]


@use_lifecycle_hooks_async(hooks=undine_settings.PARSE_HOOKS)
async def _parse_source_async(context: LifecycleHookContext) -> None:  # noqa: RUF029
    if context.result is not None:
        return

    if context.document is not None:
        return

    try:
        context.document = parse(
            source=context.source,
            no_location=undine_settings.NO_ERROR_LOCATION,
            max_tokens=undine_settings.MAX_TOKENS,
        )
    except GraphQLError as error:
        context.result = build_response(errors=[error])


@use_lifecycle_hooks_async(hooks=undine_settings.VALIDATION_HOOKS)
async def _validate_document_async(context: LifecycleHookContext) -> None:  # noqa: RUF029
    if context.result is not None:
        return

    if context.request.method != "WEBSOCKET":
        _validate_http(context)
        if context.result is not None:
            return

    validation_errors = validate(
        schema=undine_settings.SCHEMA,
        document_ast=context.document,  # type: ignore[arg-type]
        rules=get_validation_rules(),
        max_errors=undine_settings.MAX_ERRORS,
    )
    if validation_errors:
        context.result = build_response(errors=validation_errors)
        return


@use_lifecycle_hooks_async(hooks=undine_settings.EXECUTION_HOOKS)
async def _execute_async(context: LifecycleHookContext) -> ExecutionResult:
    exec_context = _get_execution_context(
        document=context.document,  # type: ignore[arg-type]
        root_value=undine_settings.ROOT_VALUE,
        context_value=context.request,
        variable_values=context.variables,
        operation_name=context.operation_name,
    )
    result = _execute(exec_context)

    if result is None:  # pragma: no cover
        raise GraphQLNoExecutionResultError

    context.result = await result if isawaitable(result) else result
    return context.result


# WebSocket execution


def raised_exceptions_as_execution_results_websocket(
    func: Callable[P, Awaitable[WebSocketResult]],
) -> Callable[P, Awaitable[WebSocketResult]]:
    @wraps(func)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> WebSocketResult:
        try:
            return await func(*args, **kwargs)

        except GraphQLError as error:
            return build_response(errors=[error])

        except GraphQLErrorGroup as error:
            return build_response(errors=list(error.flatten()))

        except Exception as error:  # noqa: BLE001
            err = GraphQLUnexpectedError(message=str(error))
            return build_response(errors=[err])

    return wrapper


@raised_exceptions_as_execution_results_websocket
async def execute_graphql_websocket(params: GraphQLHttpParams, request: DjangoRequestProtocol) -> WebSocketResult:
    """
    Executes a GraphQL operation received from an WebSocket asynchronously.
    Assumes that the schema has been validated (e.g. created using `undine.schema.create_schema`).

    :param params: GraphQL request parameters.
    :param request: The `WebSocketRequest` object to use as the GraphQL execution context value.
    """
    context = LifecycleHookContext.from_graphql_params(params=params, request=request)
    return await _run_operation_websocket(context)


@use_lifecycle_hooks_async(hooks=undine_settings.OPERATION_HOOKS)
async def _run_operation_websocket(context: LifecycleHookContext) -> WebSocketResult:
    await _parse_source_async(context)
    if context.result is None:
        await _validate_document_async(context)
        if context.result is None:
            if is_subscription_operation(context.document):  # type: ignore[arg-type]
                return await _subscribe(context)
            return await _execute_async(context)

    if context.result is None:  # pragma: no cover
        raise GraphQLNoExecutionResultError

    if isawaitable(context.result):
        context.result = await context.result  # type: ignore[assignment]

    return context.result  # type: ignore[return-value]


async def _subscribe(context: LifecycleHookContext) -> WebSocketResult:
    """Executes a subscription operation. See: `graphql.execution.subscribe.subscribe`."""
    result_or_stream = await _create_source_event_stream(context=context)
    if isinstance(result_or_stream, ExecutionResult):
        return result_or_stream

    return _map_source_to_response(source=result_or_stream, context=context)


async def _create_source_event_stream(context: LifecycleHookContext) -> AsyncIterable[Any] | ExecutionResult:
    """
    A source event stream represents a sequence of events,
    each of which triggers a GraphQL execution for that event.
    """
    context_or_errors = undine_settings.EXECUTION_CONTEXT_CLASS.build(
        schema=undine_settings.SCHEMA,
        document=context.document,  # type: ignore[arg-type]
        root_value=undine_settings.ROOT_VALUE,
        context_value=context.request,
        raw_variable_values=context.variables,
        operation_name=context.operation_name,
        middleware=undine_settings.MIDDLEWARE,
    )
    if isinstance(context_or_errors, list):
        return build_response(errors=context_or_errors)

    try:
        event_stream = await execute_subscription(context_or_errors)
    except GraphQLError as error:
        return build_response(errors=[error])

    if not isinstance(event_stream, AsyncIterable):
        err = GraphQLUnexpectedError(message="Subscription did not return an event stream")
        return build_response(errors=[err])

    return event_stream


async def _map_source_to_response(source: AsyncIterable, context: LifecycleHookContext) -> ExecutionResultGen:
    """
    For each payload yielded from a subscription,
    map it over the normal GraphQL `execute` function, with `payload` as the `root_value`.
    """
    manager = aclosing(source) if isinstance(source, AsyncGenerator) else nullcontext()

    async with manager:
        stream = aiter(source)

        while True:
            context.result = None

            async with LifecycleHookManager(hooks=undine_settings.EXECUTION_HOOKS, context=context):
                if context.result is not None:
                    yield context.result
                    continue

                try:
                    payload = await anext(stream)
                except StopAsyncIteration:
                    break

                if isinstance(payload, GraphQLError):
                    context.result = build_response(errors=[payload])
                    yield context.result
                    continue

                if isinstance(payload, GraphQLErrorGroup):
                    context.result = build_response(errors=list(payload.flatten()))
                    yield context.result
                    continue

                exec_context = _get_execution_context(
                    document=context.document,
                    root_value=payload,
                    context_value=context.request,
                    variable_values=context.variables,
                    operation_name=context.operation_name,
                )
                result = _execute(exec_context)

                context.result = await result if isawaitable(result) else result
                yield context.result


# Helpers


def _get_execution_context(
    document: DocumentNode,
    root_value: Any,
    context_value: Any,
    variable_values: dict[str, Any],
    operation_name: str | None,
) -> UndineExecutionContext:
    context_or_errors = undine_settings.EXECUTION_CONTEXT_CLASS.build(
        schema=undine_settings.SCHEMA,
        document=document,
        root_value=root_value,
        context_value=context_value,
        raw_variable_values=variable_values,
        operation_name=operation_name,
        middleware=undine_settings.MIDDLEWARE,
    )

    if isinstance(context_or_errors, list):
        raise GraphQLErrorGroup(errors=context_or_errors)

    return context_or_errors


def _execute(context: UndineExecutionContext) -> AwaitableOrValue[ExecutionResult]:
    try:
        data_or_awaitable = context.execute_operation(context.operation, context.root_value)

    except GraphQLError as error:
        context.errors.append(error)
        return context.build_response(data=None, errors=context.errors)

    except GraphQLErrorGroup as error:
        context.errors.extend(error.flatten())
        return context.build_response(data=None, errors=context.errors)

    if context.is_awaitable(data_or_awaitable):

        async def await_result() -> ExecutionResult:
            try:
                data = await data_or_awaitable

            except GraphQLError as err:
                context.errors.append(err)
                return context.build_response(data=None, errors=context.errors)

            except GraphQLErrorGroup as err:
                context.errors.extend(err.flatten())
                return context.build_response(data=None, errors=context.errors)

            else:
                return context.build_response(data=data, errors=context.errors)

        return await_result()

    return context.build_response(data=data_or_awaitable, errors=context.errors)



from __future__ import annotations

import inspect
import sys
from collections.abc import AsyncGenerator, AsyncIterable, AsyncIterator
from functools import partial
from traceback import format_tb
from types import FunctionType, GenericAlias, LambdaType, UnionType
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    NamedTuple,
    ParamSpec,
    Protocol,
    TypeGuard,
    TypeVar,
    get_args,
    get_origin,
)

from graphql import GraphQLResolveInfo

from undine.dataclasses import RootAndInfoParams
from undine.exceptions import FunctionSignatureParsingError
from undine.settings import undine_settings
from undine.typing import GQLInfo, LiteralArg, ParametrizedType

if TYPE_CHECKING:
    from collections.abc import Callable, Generator, Hashable, Sequence
    from types import FrameType, TracebackType

    from undine.typing import Lambda


__all__ = [
    "FunctionEqualityWrapper",
    "cache_signature_if_function",
    "can_be_literal_arg",
    "get_flattened_generic_params",
    "get_instance_name",
    "get_members",
    "get_root_and_info_params",
    "get_signature",
    "get_traceback",
    "get_wrapped_func",
    "has_callable_attribute",
    "is_generic_list",
    "is_lambda",
    "is_list_of",
    "is_namedtuple",
    "is_not_required_type",
    "is_protocol",
    "is_required_type",
    "is_same_func",
    "is_subclass",
    "reverse_enumerate",
]


try:  # pragma: no cover
    from typing import is_protocol
except ImportError:  # pragma: no cover

    def is_protocol(tp: type, /) -> bool:
        """Check if the given type is a Protocol."""
        return isinstance(tp, type) and getattr(tp, "_is_protocol", False) and tp != Protocol


T = TypeVar("T")
P = ParamSpec("P")
TType = TypeVar("TType", bound=type)


def get_members(obj: object, type_: type[T]) -> dict[str, T]:
    """Get members of the given object that are instances of the given type."""
    return dict(inspect.getmembers(obj, lambda x: isinstance(x, type_)))


def get_wrapped_func(func: Callable[..., Any]) -> Callable[..., Any]:
    """
    Get the inner function of a partial function, classmethod, staticmethod, property,
    or a function wrapped with `functools.wraps`.
    """
    while True:
        if hasattr(func, "__wrapped__"):  # Wrapped with functools.wraps
            func = func.__wrapped__
            continue
        if isinstance(func, partial):
            func = func.func
            continue
        if inspect.ismethod(func) and hasattr(func, "__func__"):
            func = func.__func__
            continue
        if isinstance(func, property):
            func = func.fget
            continue
        break
    return func


def get_origin_or_noop(type_hint: T) -> T:
    """
    Get the unsubscripted version of the given type hint,
    or return the type hint itself if it's not unsubscripted.
    """
    return get_origin(type_hint) or type_hint


def get_flattened_generic_params(tp: Any) -> tuple[Any, ...]:
    """
    Get all generic parameters of the given type.
    Flattens any union types.
    """
    return tuple(a for arg in get_args(tp) for a in (get_args(arg) if isinstance(arg, UnionType) else (arg,)))


def cache_signature_if_function(value: Callable[..., Any], *, depth: int = 0) -> Callable[..., Any]:
    """
    Cache signature of the given value if it's a known function type.
    This allows calling `get_signature` later without knowing the function globals or locals.

    :param value: The value to cache the signature of.
    :param depth: How many function calls deep is the code calling this method compared to the parsed function?
    :returns: The "unwrapped" function, if it was a know function type, otherwise the value as is.
    """
    value = get_wrapped_func(value)
    if isinstance(value, FunctionType):
        get_signature(value, depth=depth + 1)
    return value


class _SignatureParser:
    """
    Parse the signature of a function.

    Parsed signatures are cached so that subsequent queries for the same function
    don't need to know the globals or locals of the function scope to resolve it.
    """

    def __init__(self) -> None:
        self.cache: dict[FunctionType | Callable[..., Any], inspect.Signature] = {}

    def __call__(self, func: FunctionType | Callable[..., Any], /, *, depth: int = 0) -> inspect.Signature:
        """
        Parse the signature

        :param func: The function to parse.
        :param depth: How many function calls deep is the code calling this method compared to the parsed function?
        """
        if func in self.cache:
            return self.cache[func]

        depth += 1
        frame: FrameType = sys._getframe(depth)

        # Add some common stuff to globals so that we don't encounter as many NameErrors
        # when parsing signatures if these type hints are in a `TYPE_CHECKING` block.
        extra_globals: dict[str, Any] = {
            "AsyncIterable": AsyncIterable,
            "AsyncIterator": AsyncIterator,
            "AsyncGenerator": AsyncGenerator,
            "GraphQLResolveInfo": GraphQLResolveInfo,
            "GQLInfo": GQLInfo,
            "Any": Any,
        }

        frame_locals: dict[str, Any] = frame.f_locals
        frame_globals: dict[str, Any] = frame.f_globals | extra_globals

        # Check if previous frames are in the same file and collect their locals.
        depth += 1
        prev_frame: FrameType = sys._getframe(depth)
        while prev_frame.f_code.co_filename == frame.f_code.co_filename:
            frame_locals = prev_frame.f_locals | frame_locals  # order matters!
            depth += 1
            prev_frame = sys._getframe(depth)

        try:
            sig = inspect.signature(func, eval_str=True, globals=frame_globals, locals=frame_locals)
        except NameError as error:
            raise FunctionSignatureParsingError(name=error.name, func=func) from error

        self.cache[func] = sig
        return sig


get_signature = _SignatureParser()


def has_callable_attribute(obj: object, name: str) -> bool:
    """Check if the given object has a callable attribute with the given name."""
    return hasattr(obj, name) and callable(getattr(obj, name))


def is_subclass(obj: object, cls: TType) -> TypeGuard[TType]:
    """Check if the given object is a subclass of the given class."""
    return isinstance(obj, type) and issubclass(obj, cls)  # type: ignore[arg-type]


def is_list_of(value: Any, cls: type[T], *, allow_empty: bool = False) -> TypeGuard[list[T]]:
    """
    Check if the value is a homogeneous list of the given type.
    List must have at least one item to be considered a list of the given type, unless `allow_empty` is True.
    """
    max_length = 0 if allow_empty else 1
    return isinstance(value, list) and len(value) >= max_length and all(isinstance(item, cls) for item in value)  # type: ignore[arg-type]


def is_generic_list(type_: type) -> TypeGuard[GenericAlias]:
    """Check if the given type is a generic list, i.e., `list[str]`."""
    return isinstance(type_, GenericAlias) and issubclass(type_.__origin__, list)  # type: ignore[arg-type]


def is_lambda(func: Callable[..., Any]) -> TypeGuard[Lambda]:
    """Check if the given function is a lambda function."""
    return isinstance(func, LambdaType) and func.__name__ == "<lambda>"


def is_required_type(type_: Any) -> bool:
    """Check if the given type is a TypedDict `Required` type."""
    return isinstance(type_, ParametrizedType) and getattr(type_.__origin__, "_name", None) == "Required"  # type: ignore[misc]


def is_not_required_type(type_: Any) -> bool:
    """Check if the given type is a TypedDict `Required` type."""
    return isinstance(type_, ParametrizedType) and getattr(type_.__origin__, "_name", None) == "NotRequired"  # type: ignore[misc]


def is_namedtuple(obj: Any) -> TypeGuard[type[NamedTuple]]:
    """Check if the given object is a namedtuple class or not."""
    return (
        is_subclass(obj, tuple)
        and getattr(obj, "_fields", None) is not None
        and getattr(obj, "_field_defaults", None) is not None
        and getattr(obj, "_asdict", None) is not None
    )


def is_same_func(func_1: FunctionType | Callable[..., Any], func_2: FunctionType | Callable[..., Any], /) -> bool:
    """
    Check if the given functions are the same function.
    Handles partial functions and functions wrapped with `functools.wraps`.
    """
    return get_wrapped_func(func_1) == get_wrapped_func(func_2)


def can_be_literal_arg(key: Any) -> TypeGuard[LiteralArg]:
    return isinstance(key, LiteralArg)


def get_instance_name() -> str:
    """
    Perform some python black magic to find the name of the variable
    to which an instance of a class is being assigned to.
    Should be used in the '__init__' method.

    Note: This only works if the instance initializer is called on the
    same line as the variable for it's defined to.
    """
    frame = sys._getframe(2)
    source = inspect.findsource(frame)[0]
    line = source[frame.f_lineno - 1]
    definition = line.split("=", maxsplit=1)[0]
    return definition.split(":", maxsplit=1)[0].strip()


class FunctionEqualityWrapper(Generic[T]):
    """
    Adds equality checks for a function based on the provided context.
    Function is equal to another function if it's also wrapped with this class
    and the provided contexts are equal.
    """

    def __init__(self, func: Callable[[], T], context: Hashable) -> None:
        self.func = func
        self.context = context

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.context == other.context

    def __hash__(self) -> int:
        return hash(self.context)

    def __call__(self) -> T:
        return self.func()


def get_root_and_info_params(func: FunctionType | Callable[..., Any], *, depth: int = 0) -> RootAndInfoParams:
    """
    Inspect the function signature to figure out which parameters are
    the root and info parameters of a GraphQL resolver function, if any.

    `root_param` is the first parameter of the function if it's named `self`, `cls`,
    or the name configured with the `RESOLVER_ROOT_PARAM_NAME` setting (`root` by default).

    `info_param` is annotated as `GraphQLResolveInfo` or `GQLInfo` if it exists,
    and is usually the second parameter of the function, but can also be in any
    other position other than first.
    """
    sig = get_signature(func, depth=depth + 1)

    root_param: str | None = None
    info_param: str | None = None
    for i, param in enumerate(sig.parameters.values()):
        if i == 0 and param.name in {"self", "cls", undine_settings.RESOLVER_ROOT_PARAM_NAME}:
            root_param = param.name

        elif get_origin_or_noop(param.annotation) in {GraphQLResolveInfo, GQLInfo}:
            info_param = param.name
            break

    return RootAndInfoParams(root_param=root_param, info_param=info_param)


def reverse_enumerate(sequence: Sequence[T]) -> Generator[tuple[int, T], None, None]:
    """
    Enumerate the given sequence in reverse order.
    Using this allows using `.pop(index)` on the sequence to remove the iterated item if needed.

    >>> x = ["a", "b", "c"]
    >>> for i, item in reverse_enumerate(x):
    >>>     x.pop(i)
    """
    for index in range(len(sequence) - 1, -1, -1):
        yield index, sequence[index]


def get_traceback(traceback: TracebackType) -> list[str]:
    """Format the given traceback into a list of strings."""
    return [subline for line in format_tb(traceback) for subline in line.split("\n")]



from __future__ import annotations

import dataclasses
from collections import defaultdict
from copy import deepcopy
from typing import TYPE_CHECKING, Any, Literal, TypeAlias, overload

from django.db.models import Model
from django.db.models.manager import BaseManager
from graphql import Undefined

from undine.exceptions import (
    GraphQLInvalidInputDataError,
    GraphQLMutationInputNotFoundError,
    GraphQLMutationInstanceLimitError,
)
from undine.parsers import parse_model_relation_info
from undine.settings import undine_settings
from undine.typing import RelatedAction

from .model_utils import generic_relations_for_generic_foreign_key, get_instances_or_raise
from .reflection import is_subclass
from .text import to_camel_case

if TYPE_CHECKING:
    from collections.abc import Iterator

    from django.contrib.contenttypes.fields import GenericForeignKey

    from undine import MutationType
    from undine.dataclasses import RelInfo

__all__ = [
    "MutationData",
    "MutationManyData",
    "get_mutation_data",
]


Placements: TypeAlias = defaultdict[type[Model], dict[tuple[str | int, ...], Any]]


@overload
def get_mutation_data(
    *,
    model: type[Model],
    data: dict[str, Any],
    mutation_type: type[MutationType] | None = None,
) -> MutationData: ...


@overload
def get_mutation_data(
    *,
    model: type[Model],
    data: list[dict[str, Any]],
    mutation_type: type[MutationType] | None = None,
) -> list[MutationData]: ...


def get_mutation_data(
    *,
    model: type[Model],
    data: dict[str, Any] | list[dict[str, Any]],
    mutation_type: type[MutationType] | None = None,
) -> MutationData | list[MutationData]:
    if single := isinstance(data, dict):
        data = [data]

    mutation_data = build_mutation_data(model=model, data=data, mutation_type=mutation_type)

    if single:
        return mutation_data[0]
    return mutation_data


def build_mutation_data(
    *,
    model: type[Model],
    data: list[dict[str, Any]],
    mutation_type: type[MutationType] | None = None,
) -> list[MutationData]:
    mutation_infos = [get_mutation_info(model=model, data=item) for item in data]

    instance_count = sum(mutation_info.instance_count for mutation_info in mutation_infos)
    if instance_count > undine_settings.MUTATION_INSTANCE_LIMIT:
        raise GraphQLMutationInstanceLimitError(limit=undine_settings.MUTATION_INSTANCE_LIMIT, count=instance_count)

    placements: Placements = defaultdict(dict)

    for i, mutation_info in enumerate(mutation_infos):
        get_placements(placements, data=mutation_info, path=[i])

    for path_model, path_to_pk in placements.items():
        pks = set(path_to_pk.values())

        # TODO: Could join required relations. Should still check that fetched instance count is correct.
        instances = get_instances_or_raise(model=path_model, pks=pks)
        instances_map = {instance.pk: instance for instance in instances}

        for path, pk in path_to_pk.items():
            obj: Any = mutation_infos
            for key in path:
                obj = obj[key]

            mut_info: MutationInfo = obj
            mut_info.instance = instances_map[pk]

    return [mutation_info.process(mutation_type) for mutation_info in mutation_infos]


def get_placements(placements: Placements, *, data: MutationInfo, path: list[str | int]) -> None:
    if "pk" in data.fields:
        placements[data.model][tuple(path)] = data.fields["pk"]

    for key, single_data in data.single_relations.items():
        if single_data is not None:
            get_placements(placements, data=single_data, path=[*path, "single_relations", key])

    for key, many_data in data.many_relations.items():
        for i, item_data in enumerate(many_data):
            get_placements(placements, data=item_data, path=[*path, "many_relations", key, i])


def get_mutation_info(*, model: type[Model], data: dict[str, Any]) -> MutationInfo:
    relation_info = parse_model_relation_info(model=model)

    fields: dict[str, Any] = {}
    single_relations: dict[str, MutationInfo | None] = {}
    many_relations: dict[str, list[MutationInfo]] = {}

    for field_name, field_data in data.items():
        rel_info = relation_info.get(field_name)
        if rel_info is None:
            fields[field_name] = field_data
            continue

        # Generic foreign key
        if rel_info.relation_type.is_generic_foreign_key:
            single_relations[field_name] = get_generic_foreign_key_info(data=field_data, rel_info=rel_info)
            continue

        # Single relations
        if rel_info.relation_type.is_single:
            single_relations[field_name] = get_related_info(data=field_data, rel_info=rel_info)
            continue

        # Many relations
        values = many_relations.setdefault(field_name, [])
        for item_data in field_data or []:  # 'field_data' can be None
            mutation_info = get_related_info(data=item_data, rel_info=rel_info)
            if mutation_info is not None:
                values.append(mutation_info)

    return MutationInfo(
        instance=model(),
        model=model,
        fields=fields,
        single_relations=single_relations,
        many_relations=many_relations,
    )


def get_related_info(*, data: Any, rel_info: RelInfo) -> MutationInfo | None:
    match data:
        case dict():
            return get_mutation_info(model=rel_info.related_model, data=data)  # type: ignore[arg-type]

        case rel_info.related_model_pk_type():
            return MutationInfo(
                instance=rel_info.related_model(),  # type: ignore[misc]
                model=rel_info.related_model,  # type: ignore[arg-type]
                fields={"pk": data},
            )

        case rel_info.related_model():
            return MutationInfo(
                instance=data,
                model=rel_info.related_model,  # type: ignore[arg-type]
            )

        case None:
            # TODO: Could validate if relation is nullable?
            return None

    raise GraphQLInvalidInputDataError(field_name=rel_info.field_name, data=data)


def get_generic_foreign_key_info(*, data: Any, rel_info: RelInfo) -> MutationInfo | None:
    if not isinstance(data, dict):
        raise GraphQLInvalidInputDataError(field_name=rel_info.field_name, data=data)

    key: str | None = next(iter(data), None)
    if key is None:
        raise GraphQLInvalidInputDataError(field_name=rel_info.field_name, data=data)

    model_data = data[key]

    field: GenericForeignKey = rel_info.model._meta.get_field(rel_info.field_name)  # type: ignore[assignment]
    relations = generic_relations_for_generic_foreign_key(field)
    related_model_map = {to_camel_case(rel.model.__name__): rel.model for rel in relations}

    model = related_model_map.get(key)
    if model is None:
        msg = f"Model '{key}' doesn't exist or have a generic relation to '{rel_info.model.__name__}'."
        raise GraphQLInvalidInputDataError(msg)

    match model_data:
        case dict():
            return get_mutation_info(model=model, data=model_data)

        case None:
            # TODO: Could validate if relation is nullable?
            return None

    raise GraphQLInvalidInputDataError(field_name=rel_info.field_name, data=model_data)


@dataclasses.dataclass(kw_only=True)
class MutationInfo:
    instance: Model
    model: type[Model]
    fields: dict[str, Any] = dataclasses.field(default_factory=dict)
    single_relations: dict[str, MutationInfo | None] = dataclasses.field(default_factory=dict)
    many_relations: dict[str, list[MutationInfo]] = dataclasses.field(default_factory=dict)

    def __getitem__(self, key: Literal["single_relations", "many_relations"]) -> Any:
        # For setting instances
        try:
            return getattr(self, key)
        except AttributeError as error:
            raise KeyError(key) from error

    @property
    def instance_count(self) -> int:
        instance_count = 1

        for single_data in self.single_relations.values():
            if single_data is not None:
                instance_count += single_data.instance_count

        for many_data in self.many_relations.values():
            for item_data in many_data:
                instance_count += item_data.instance_count

        return instance_count

    def process(self, mutation_type: type[MutationType] | None) -> MutationData:
        data = deepcopy(self.fields)

        for key, single_data in self.single_relations.items():
            sub_mutation_type = get_nested_mutation_type(mutation_type=mutation_type, field_name=key)
            data[key] = single_data.process(sub_mutation_type) if single_data is not None else None

        for key, many_data in self.many_relations.items():
            sub_mutation_type = get_nested_mutation_type(mutation_type=mutation_type, field_name=key)
            values = data.setdefault(key, MutationManyData(mutation_type=sub_mutation_type))
            for item_data in many_data:
                values.append(item_data.process(sub_mutation_type))

        return MutationData(instance=self.instance, data=data, mutation_type=mutation_type)


@dataclasses.dataclass(kw_only=True)
class MutationData:
    """Processed data that everything needed to mutate the given instance."""

    instance: Model
    """The instance that the data will be applied to."""

    data: dict[str, Any]  # Node: `Any` can be `MutationData` or `MutationManyData`
    """The data that will be applied to the instance as part of this mutation."""

    mutation_type: type[MutationType] | None
    """If this is a related mutation, what action should be taken on non-updated related objects."""

    @property
    def related_action(self) -> RelatedAction:
        return self.mutation_type.__related_action__ if self.mutation_type is not None else RelatedAction.null

    @property
    def plain_data(self) -> dict[str, Any]:
        """Plain data without additional information."""
        data: dict[str, Any] = {}

        for key, value in self.data.items():
            if isinstance(value, MutationData):
                if value.mutation_type is not None:
                    data[key] = value.plain_data
                    continue

                if "pk" in value.data and len(value.data) == 1:
                    data[key] = value.data["pk"]
                    continue

                if value.instance.pk is not None:
                    data[key] = value.instance
                    continue

                data[key] = value.plain_data
                continue

            if isinstance(value, MutationManyData):
                if all(item.mutation_type is not None for item in value):
                    data[key] = [item.plain_data for item in value]
                    continue

                if all(("pk" in item.data and len(item.data) == 1) for item in value):
                    data[key] = [item.data["pk"] for item in value]
                    continue

                if all(item.instance.pk is not None for item in value):
                    data[key] = [item.instance for item in value]
                    continue

                data[key] = [item.plain_data for item in value]
                continue

            data[key] = value

        return data

    @property
    def previous_data(self) -> dict[str, Any]:
        """What data will will change in the given instance if this mutation is performed."""
        if self.instance.pk is None:
            return {}

        data: dict[str, Any] = {}
        for key, value in self.data.items():
            if isinstance(value, MutationData):
                data[key] = value.previous_data
                continue

            if isinstance(value, MutationManyData):
                data[key] = [item.previous_data for item in value]
                continue

            previous_value = getattr(self.instance, key, Undefined)
            if previous_value == Undefined:
                continue

            if isinstance(previous_value, BaseManager):
                previous_value = list(previous_value.all())

            data[key] = previous_value

        return data


@dataclasses.dataclass(kw_only=True)
class MutationManyData:
    """A `MutationData` that can be used for many related fields."""

    mutation_data: list[MutationData] = dataclasses.field(default_factory=list)

    mutation_type: type[MutationType] | None
    """If this is a related mutation, what action should be taken on non-updated related objects."""

    @property
    def related_action(self) -> RelatedAction:
        return self.mutation_type.__related_action__ if self.mutation_type is not None else RelatedAction.null

    @property
    def plain_data(self) -> list[dict[str, Any]]:
        return [item.plain_data for item in self.mutation_data]

    @property
    def previous_data(self) -> list[dict[str, Any]]:
        return [item.previous_data for item in self.mutation_data]

    def __len__(self) -> int:
        return len(self.mutation_data)

    def __iter__(self) -> Iterator[MutationData]:
        return iter(self.mutation_data)

    def __getitem__(self, index: int) -> MutationData:
        return self.mutation_data[index]

    def __setitem__(self, index: int, value: MutationData) -> None:
        self.mutation_data[index] = value

    def append(self, value: MutationData) -> None:
        self.mutation_data.append(value)


def get_nested_mutation_type(mutation_type: type[MutationType] | None, field_name: str) -> type[MutationType] | None:
    if mutation_type is None:
        return None

    input_field = mutation_type.__input_map__.get(field_name)

    if input_field is None:  # pragma: no cover
        raise GraphQLMutationInputNotFoundError(field_name=field_name, mutation_type=mutation_type)

    from undine import MutationType  # noqa: PLC0415

    if is_subclass(input_field.ref, MutationType):
        return input_field.ref

    return None



from __future__ import annotations

import datetime
import decimal
import uuid
from contextlib import suppress
from decimal import Decimal
from types import FunctionType, NoneType
from typing import Any

from django.contrib.contenttypes.fields import GenericForeignKey, GenericRel, GenericRelation
from django.db.models import (
    BigIntegerField,
    BinaryField,
    BooleanField,
    CharField,
    DateField,
    DateTimeField,
    DecimalField,
    DurationField,
    FileField,
    FloatField,
    ImageField,
    IntegerField,
    JSONField,
    Q,
    TextField,
    TimeField,
    UUIDField,
)
from django.db.models.fields.related_descriptors import (
    ForwardManyToOneDescriptor,
    ManyToManyDescriptor,
    ReverseManyToOneDescriptor,
    ReverseOneToOneDescriptor,
)
from django.db.models.query_utils import DeferredAttribute
from graphql import GraphQLList, GraphQLNonNull, GraphQLScalarType

from undine.converters import convert_to_python_type
from undine.exceptions import FunctionDispatcherError
from undine.parsers import parse_first_param_type, parse_return_annotation
from undine.scalars import (
    GraphQLAny,
    GraphQLBase16,
    GraphQLBase32,
    GraphQLBase64,
    GraphQLBoolean,
    GraphQLDate,
    GraphQLDateTime,
    GraphQLDecimal,
    GraphQLDuration,
    GraphQLEmail,
    GraphQLFile,
    GraphQLFloat,
    GraphQLID,
    GraphQLImage,
    GraphQLInt,
    GraphQLJSON,
    GraphQLNull,
    GraphQLString,
    GraphQLTime,
    GraphQLURL,
    GraphQLUUID,
)
from undine.typing import CombinableExpression, ToManyField, ToOneField
from undine.utils.model_fields import TextChoicesField


@convert_to_python_type.register
def _(_: CharField | TextField, **kwargs: Any) -> type:
    # CharField might have an enum, but we cannot access it anymore.
    return str


@convert_to_python_type.register
def _(ref: TextChoicesField, **kwargs: Any) -> type:
    return ref.choices_enum


@convert_to_python_type.register
def _(_: BooleanField, **kwargs: Any) -> type:
    return bool


@convert_to_python_type.register
def _(_: IntegerField | BigIntegerField, **kwargs: Any) -> type:
    return int


@convert_to_python_type.register
def _(_: FloatField, **kwargs: Any) -> type:
    return float


@convert_to_python_type.register
def _(_: DecimalField, **kwargs: Any) -> type:
    return Decimal


@convert_to_python_type.register
def _(_: DateField, **kwargs: Any) -> type:
    return datetime.date


@convert_to_python_type.register
def _(_: DateTimeField, **kwargs: Any) -> type:
    return datetime.datetime


@convert_to_python_type.register
def _(_: TimeField, **kwargs: Any) -> type:
    return datetime.time


@convert_to_python_type.register
def _(_: DurationField, **kwargs: Any) -> type:
    return datetime.timedelta


@convert_to_python_type.register
def _(_: BinaryField, **kwargs: Any) -> type:
    return bytes


@convert_to_python_type.register
def _(_: UUIDField, **kwargs: Any) -> type:
    return uuid.UUID


@convert_to_python_type.register
def _(_: FileField, **kwargs: Any) -> type:
    return str


@convert_to_python_type.register
def _(_: ImageField, **kwargs: Any) -> type:
    return str


@convert_to_python_type.register
def _(_: JSONField, **kwargs: Any) -> type:
    return dict[str, str]


@convert_to_python_type.register
def _(ref: ToManyField, **kwargs: Any) -> type:
    generic_type = convert_to_python_type(ref.target_field, **kwargs)
    return list.__class_getitem__(generic_type)


@convert_to_python_type.register
def _(ref: ToOneField, **kwargs: Any) -> type:
    return convert_to_python_type(ref.target_field, **kwargs)


@convert_to_python_type.register
def _(_: Q, **kwargs: Any) -> type:
    return bool


@convert_to_python_type.register
def _(ref: CombinableExpression, **kwargs: Any) -> type:
    return convert_to_python_type(ref.output_field, **kwargs)


@convert_to_python_type.register
def _(ref: DeferredAttribute | ForwardManyToOneDescriptor, **kwargs: Any) -> type:
    return convert_to_python_type(ref.field, **kwargs)


@convert_to_python_type.register
def _(ref: ReverseManyToOneDescriptor, **kwargs: Any) -> type:
    return convert_to_python_type(ref.rel, **kwargs)


@convert_to_python_type.register
def _(ref: ReverseOneToOneDescriptor, **kwargs: Any) -> type:
    return convert_to_python_type(ref.related, **kwargs)


@convert_to_python_type.register
def _(ref: ManyToManyDescriptor, **kwargs: Any) -> type:
    return convert_to_python_type(ref.rel if ref.reverse else ref.field, **kwargs)


@convert_to_python_type.register
def _(ref: GraphQLNonNull, **kwargs: Any) -> type:
    return convert_to_python_type(ref.of_type, **kwargs)


@convert_to_python_type.register
def _(ref: GraphQLList, **kwargs: Any) -> type:
    item_type = convert_to_python_type(ref.of_type, **kwargs)
    return list.__class_getitem__(item_type)


@convert_to_python_type.register
def _(ref: FunctionType, **kwargs: Any) -> type:
    is_input = kwargs.get("is_input", False)
    return parse_first_param_type(ref) if is_input else parse_return_annotation(ref)


@convert_to_python_type.register
def _(ref: GenericRelation, **kwargs: Any) -> type:
    generic_type = convert_to_python_type(ref.target_field, **kwargs)
    return list.__class_getitem__(generic_type)


@convert_to_python_type.register
def _(ref: GenericRel, **kwargs: Any) -> type:
    return convert_to_python_type(ref.field)


@convert_to_python_type.register
def _(ref: GenericForeignKey, **kwargs: Any) -> type:
    field = ref.model._meta.get_field(ref.fk_field)
    return convert_to_python_type(field)


@convert_to_python_type.register
def _(ref: GraphQLScalarType, **kwargs: Any) -> type:  # noqa: PLR0911, PLR0912
    match ref.name:
        case GraphQLID.name | GraphQLString.name | GraphQLEmail.name | GraphQLURL.name:
            return str
        case GraphQLBoolean.name:
            return bool
        case GraphQLInt.name:
            return int
        case GraphQLFloat.name:
            return float
        case GraphQLDecimal.name:
            return decimal.Decimal
        case GraphQLDate.name:
            return datetime.date
        case GraphQLTime.name:
            return datetime.time
        case GraphQLDateTime.name:
            return datetime.datetime
        case GraphQLDuration.name:
            return datetime.timedelta
        case GraphQLUUID.name:
            return uuid.UUID
        case GraphQLAny.name:
            return Any
        case GraphQLNull.name:
            return NoneType
        case GraphQLJSON.name:
            return dict
        case GraphQLBase16.name | GraphQLBase32.name | GraphQLBase64.name:
            return bytes
        case GraphQLFile.name | GraphQLImage.name:
            return str

    msg = f"Unknown GraphQLScalarType: '{ref.name}'. Cannot find matching python type."
    raise FunctionDispatcherError(msg)


with suppress(ImportError):
    from django.contrib.postgres.fields import ArrayField, HStoreField

    @convert_to_python_type.register
    def _(ref: HStoreField, **kwargs: Any) -> type:
        return dict[str, str]

    @convert_to_python_type.register
    def _(ref: ArrayField, **kwargs: Any) -> type:
        item_type = convert_to_python_type(ref.base_field, **kwargs)
        return list.__class_getitem__(item_type)


with suppress(ImportError):
    from django.db.models import GeneratedField

    @convert_to_python_type.register
    def _(ref: GeneratedField, **kwargs: Any) -> type:
        return convert_to_python_type(ref.output_field, **kwargs)



from __future__ import annotations

import datetime
import decimal
import uuid
from enum import Enum
from types import FunctionType
from typing import Any

from django.contrib.contenttypes.fields import GenericForeignKey, GenericRel, GenericRelation
from django.db.models import F, Model, TextChoices
from django.db.models.fields.related_descriptors import (
    ForwardManyToOneDescriptor,
    ManyToManyDescriptor,
    ReverseManyToOneDescriptor,
    ReverseOneToOneDescriptor,
)
from django.db.models.query_utils import DeferredAttribute

from undine import Input, MutationType
from undine.converters import convert_to_input_ref, is_many
from undine.dataclasses import LazyLambda, TypeRef
from undine.exceptions import InvalidInputMutationTypeError, ModelFieldDoesNotExistError
from undine.settings import undine_settings
from undine.typing import GenericField, Lambda, ModelField, MutationKind, RelatedField, RelationType
from undine.utils.model_utils import (
    generic_foreign_key_for_generic_relation,
    get_instance_or_raise,
    get_instances_or_raise,
    get_model_field,
    get_related_name,
)


@convert_to_input_ref.register
def _(_: None, **kwargs: Any) -> Any:
    caller: Input = kwargs["caller"]
    field = get_model_field(model=caller.mutation_type.__model__, lookup=caller.field_name)
    return convert_to_input_ref(field, **kwargs)


@convert_to_input_ref.register
def _(ref: F, **kwargs: Any) -> Any:
    caller: Input = kwargs["caller"]
    field = get_model_field(model=caller.mutation_type.__model__, lookup=ref.name)
    return convert_to_input_ref(field, **kwargs)


@convert_to_input_ref.register
def _(ref: ModelField, **kwargs: Any) -> Any:
    return ref


@convert_to_input_ref.register
def _(ref: type[Model], **kwargs: Any) -> Any:
    caller: Input = kwargs["caller"]

    if undine_settings.ASYNC:
        msg = "Model Inputs cannot be used in async mutations."
        raise RuntimeError(msg)  # TODO: Custom exception

    user_func = caller.convertion_func

    if is_many(ref, model=caller.mutation_type.__model__, name=caller.field_name):

        def convert_many(inpt: Input, value: list[Any] | None) -> list[Model]:
            instances: list[Model] = []
            if value is not None:
                instances = get_instances_or_raise(model=ref, pks=set(value))
            if user_func is not None:
                return user_func(inpt, instances)
            return instances

        caller.convertion_func = convert_many
        return ref

    def convert_single(inpt: Input, value: Any) -> Model | None:
        instance: Model | None = None
        if value is not None:
            instance = get_instance_or_raise(model=ref, pk=value)
        if user_func is not None:
            return user_func(inpt, instance)
        return instance

    caller.convertion_func = convert_single
    return ref


@convert_to_input_ref.register
def _(ref: DeferredAttribute | ForwardManyToOneDescriptor, **kwargs: Any) -> Any:
    return ref.field


@convert_to_input_ref.register
def _(ref: ReverseManyToOneDescriptor, **kwargs: Any) -> Any:
    return convert_to_input_ref(ref.rel, **kwargs)


@convert_to_input_ref.register
def _(ref: ReverseOneToOneDescriptor, **kwargs: Any) -> Any:
    return convert_to_input_ref(ref.related, **kwargs)


@convert_to_input_ref.register
def _(ref: ManyToManyDescriptor, **kwargs: Any) -> Any:
    return convert_to_input_ref(ref.rel if ref.reverse else ref.field, **kwargs)


@convert_to_input_ref.register
def _(ref: str, **kwargs: Any) -> Any:
    caller: Input = kwargs["caller"]
    field = get_model_field(model=caller.mutation_type.__model__, lookup=ref)
    return convert_to_input_ref(field, **kwargs)


@convert_to_input_ref.register
def _(ref: Lambda, **kwargs: Any) -> Any:
    return LazyLambda(callback=ref)


@convert_to_input_ref.register
def _(ref: type[str], **kwargs: Any) -> Any:
    return TypeRef(value=ref)


@convert_to_input_ref.register
def _(ref: type[bool], **kwargs: Any) -> Any:
    return TypeRef(value=ref)


@convert_to_input_ref.register
def _(ref: type[int], **kwargs: Any) -> Any:
    return TypeRef(value=ref)


@convert_to_input_ref.register
def _(ref: type[float], **kwargs: Any) -> Any:
    return TypeRef(value=ref)


@convert_to_input_ref.register
def _(ref: type[decimal.Decimal], **kwargs: Any) -> Any:
    return TypeRef(value=ref)


@convert_to_input_ref.register
def _(ref: type[datetime.datetime], **kwargs: Any) -> Any:
    return TypeRef(value=ref)


@convert_to_input_ref.register
def _(ref: type[datetime.date], **kwargs: Any) -> Any:
    return TypeRef(value=ref)


@convert_to_input_ref.register
def _(ref: type[datetime.time], **kwargs: Any) -> Any:
    return TypeRef(value=ref)


@convert_to_input_ref.register
def _(ref: type[datetime.timedelta], **kwargs: Any) -> Any:
    return TypeRef(value=ref)


@convert_to_input_ref.register
def _(ref: type[uuid.UUID], **kwargs: Any) -> Any:
    return TypeRef(value=ref)


@convert_to_input_ref.register
def _(ref: type[Enum], **kwargs: Any) -> Any:
    return TypeRef(value=ref)


@convert_to_input_ref.register
def _(ref: type[TextChoices], **kwargs: Any) -> Any:
    return TypeRef(value=ref)


@convert_to_input_ref.register
def _(ref: type[list], **kwargs: Any) -> Any:
    return TypeRef(value=ref)


@convert_to_input_ref.register
def _(ref: type[dict], **kwargs: Any) -> Any:
    return TypeRef(value=ref)


@convert_to_input_ref.register
def _(ref: FunctionType, **kwargs: Any) -> Any:
    return ref


@convert_to_input_ref.register
def _(ref: type[MutationType], **kwargs: Any) -> Any:
    if ref.__kind__ != MutationKind.related:
        raise InvalidInputMutationTypeError(ref=ref, kind=ref.__kind__)

    caller: Input = kwargs["caller"]

    model = caller.mutation_type.__model__

    try:
        field: RelatedField | GenericField = get_model_field(model=model, lookup=caller.field_name)  # type: ignore[assignment]

    except ModelFieldDoesNotExistError:
        # Allow using related inputs for non-relational fields in custom mutations.
        if caller.mutation_type.__kind__ == MutationKind.custom:
            return ref
        raise

    relation_type = RelationType.for_related_field(field)

    # Remove the Input for the reverse relation from the MutationType used as the Input for this one.
    if relation_type.is_generic_relation:
        reverse_field_name: str = generic_foreign_key_for_generic_relation(field).name  # type: ignore[arg-type]
        ref.__input_map__.pop(reverse_field_name, None)

    elif not relation_type.is_generic_foreign_key:
        reverse_field_name = get_related_name(field.remote_field)  # type: ignore[arg-type]
        ref.__input_map__.pop(reverse_field_name, None)

    return ref


@convert_to_input_ref.register
def _(ref: GenericRelation, **kwargs: Any) -> Any:
    return ref


@convert_to_input_ref.register
def _(ref: GenericRel, **kwargs: Any) -> Any:
    return ref.field


@convert_to_input_ref.register  # Required for Django<5.1
def _(ref: GenericForeignKey, **kwargs: Any) -> Any:
    return ref



from __future__ import annotations

import base64
from copy import copy
from typing import TYPE_CHECKING

from django.db.models import F, ManyToManyField, ManyToManyRel, OuterRef, Value, Window
from django.db.models.functions import Greatest, RowNumber
from graphql import GraphQLBoolean, GraphQLField, GraphQLID, GraphQLNonNull, GraphQLString
from graphql.type.scalars import serialize_id

from undine.dataclasses import ValidatedPaginationArgs
from undine.exceptions import GraphQLPaginationArgumentValidationError
from undine.interface import InterfaceField, InterfaceType
from undine.optimizer.prefetch_hack import register_for_prefetch_hack
from undine.settings import undine_settings
from undine.utils.graphql.type_registry import get_or_create_graphql_object_type
from undine.utils.model_utils import SubqueryCount

if TYPE_CHECKING:
    from django.db.models import QuerySet

    from undine.query import QueryType
    from undine.typing import CombinableExpression, GQLInfo, ToManyField

__all__ = [
    "Connection",
    "Node",
    "PageInfoType",
    "PaginationHandler",
    "cursor_to_offset",
    "decode_base64",
    "encode_base64",
    "from_global_id",
    "offset_to_cursor",
    "to_global_id",
]


class PaginationHandler:
    """Handles pagination for Relay Connection based on the given arguments."""

    def __init__(
        self,
        *,
        typename: str,
        after: str | None = None,
        before: str | None = None,
        first: int | None = None,
        last: int | None = None,
        offset: int | None = None,
        page_size: int | None = None,
    ) -> None:
        """
        Create a new PaginationHandler.

        :param typename: The typename of the node in the connection.
        :param first: Number of item to return from the start.
        :param last: Number of item to return from the end (after applying `first`).
        :param after: Cursor value for the last item in the previous page.
        :param before: Cursor value for the first item in the next page.
        :param offset: Number of item to skip from the start.
        :param page_size: Maximum limit for the number of item that can be requested in a page. No limit if `None`.
        """
        validated_args = self.validate(
            typename=typename,
            first=first,
            last=last,
            offset=offset,
            after_cursor=after,
            before_cursor=before,
            page_size=page_size,
        )

        self.after = validated_args.after
        """The index after which to start (exclusive)."""

        self.before = validated_args.before
        """The index before which to stop (exclusive)."""

        self.first = validated_args.first
        """The number of items to return from the start."""

        self.last = validated_args.last
        """The number of items to return from the end (after evaluating first)."""

        self.page_size = page_size
        """Maximum number of item to return in a page. No limit if `None`."""

        # Calculated in `paginate_queryset` or `paginate_prefetch_queryset` if needed.
        self.start: int = 0
        """The index to start the pagination from."""

        # Calculated in `paginate_queryset` or `paginate_prefetch_queryset` if needed.
        self.stop: int | None = None
        """The index to stop the pagination at."""

        # Modified in `paginate_queryset` or `paginate_prefetch_queryset` if needed.
        self.total_count: int | None = None
        """The total number of items that can be paginated."""

        # Modified during optimization based on connection params.
        self.requires_total_count: bool = False
        """Whether the total count is required for this query."""

    @staticmethod
    def validate(  # noqa: C901, PLR0912
        *,
        typename: str,
        first: int | None,
        last: int | None,
        offset: int | None,
        after_cursor: str | None,
        before_cursor: str | None,
        page_size: int | None,
    ) -> ValidatedPaginationArgs:
        """Validate the given pagination arguments and return the validated arguments."""
        try:
            after = cursor_to_offset(typename, after_cursor) if after_cursor is not None else None
        except Exception as error:
            msg = f"Argument 'after' is not a valid cursor for type '{typename}'."
            raise GraphQLPaginationArgumentValidationError(msg) from error

        try:
            before = cursor_to_offset(typename, before_cursor) if before_cursor is not None else None
        except Exception as error:
            msg = f"Argument 'before' is not a valid cursor for type '{typename}'."
            raise GraphQLPaginationArgumentValidationError(msg) from error

        if page_size is not None and (not isinstance(page_size, int) or page_size < 1):
            msg = f"`page_size` must be `None` or a positive integer, got: {page_size!r}"
            raise GraphQLPaginationArgumentValidationError(msg)

        if first is not None:
            if not isinstance(first, int) or first <= 0:
                msg = "Argument 'first' must be a positive integer."
                raise GraphQLPaginationArgumentValidationError(msg)

            if isinstance(page_size, int) and first > page_size:
                msg = f"Requesting first {first} records exceeds the limit of {page_size}."
                raise GraphQLPaginationArgumentValidationError(msg)

        if last is not None:
            if not isinstance(last, int) or last <= 0:
                msg = "Argument 'last' must be a positive integer."
                raise GraphQLPaginationArgumentValidationError(msg)

            if isinstance(page_size, int) and last > page_size:
                msg = f"Requesting last {last} records exceeds the limit of {page_size}."
                raise GraphQLPaginationArgumentValidationError(msg)

        if isinstance(page_size, int) and first is None and last is None:
            first = page_size

        if offset is not None:
            if after is not None or before is not None:
                msg = "Can only use either `offset` or `before`/`after` for pagination."
                raise GraphQLPaginationArgumentValidationError(msg)
            if not isinstance(offset, int) or offset < 0:
                msg = "Argument `offset` must be a positive integer."
                raise GraphQLPaginationArgumentValidationError(msg)

            # Convert offset to after cursor value. Note that after cursor dictates
            # a value _after_ which results should be returned, so we need to subtract
            # 1 from the offset to get the correct cursor value.
            if offset > 0:  # ignore zero offset
                after = offset - 1

        if after is not None and (not isinstance(after, int) or after < 0):
            msg = "The node pointed with `after` does not exist."
            raise GraphQLPaginationArgumentValidationError(msg)

        if before is not None and (not isinstance(before, int) or before < 0):
            msg = "The node pointed with `before` does not exist."
            raise GraphQLPaginationArgumentValidationError(msg)

        if after is not None and before is not None and after >= before:
            msg = "The node pointed with `after` must be before the node pointed with `before`."
            raise GraphQLPaginationArgumentValidationError(msg)

        # Since `after` is also exclusive, we need to add 1 to it, so that slicing works correctly.
        if after is not None:
            after += 1

        return ValidatedPaginationArgs(after=after, before=before, first=first, last=last)

    def paginate_queryset(self, queryset: QuerySet, info: GQLInfo) -> QuerySet:
        """Paginate a top-level queryset."""
        self.calculate_pagination_arguments(queryset, info)
        return self.apply_pagination(queryset, info)

    def calculate_pagination_arguments(self, queryset: QuerySet, info: GQLInfo) -> None:
        """
        Calculates the pagination arguments for a top-level queryset.

        This function is based on the Relay pagination algorithm.
        See. https://relay.dev/graphql/connections.htm#sec-Pagination-algorithm
        """
        if self.requires_total_count:
            self.total_count = queryset.count()

        if self.after is not None:
            self.start = self.after

        if self.before is not None:
            self.stop = self.before

        if self.first is not None:
            self.stop = self.start + self.first if self.stop is None else min(self.start + self.first, self.stop)

        if self.last is not None:
            if self.stop is None:
                if self.total_count is None:
                    self.total_count = queryset.count()
                self.stop = self.total_count
            self.start = max(self.stop - self.last, self.start)

    def apply_pagination(self, queryset: QuerySet, info: GQLInfo) -> QuerySet:
        """Paginate a top-level queryset using queryset slicing."""
        return queryset[self.start : self.stop]

    def paginate_prefetch_queryset(self, queryset: QuerySet, field: ToManyField, info: GQLInfo) -> QuerySet:
        """Paginate a prefetch queryset."""
        self.calculate_prefetch_pagination_arguments(queryset, field, info)
        return self.apply_prefetch_pagination(queryset, field, info)

    def calculate_prefetch_pagination_arguments(self, queryset: QuerySet, field: ToManyField, info: GQLInfo) -> None:
        """
        Calculates the pagination arguments for a prefetch queryset.

        This function is based on the Relay pagination algorithm.
        See. https://relay.dev/graphql/connections.htm#sec-Pagination-algorithm
        """
        if self.requires_total_count:
            self.total_count = F(undine_settings.CONNECTION_TOTAL_COUNT_KEY)

        if self.after is not None:
            self.start = self.after

        if self.before is not None:
            self.stop = self.before

        if self.first is not None:
            self.stop = self.start + self.first if self.stop is None else min(self.start + self.first, self.stop)

        if self.last is not None:
            if self.stop is None:
                if self.total_count is None:
                    self.total_count = F(undine_settings.CONNECTION_TOTAL_COUNT_KEY)  # type: ignore[assignment]
                self.start = Greatest(self.total_count - Value(self.last), Value(self.start))  # type: ignore[assignment]
            else:
                self.start = max(self.stop - self.last, self.start)

    def apply_prefetch_pagination(self, queryset: QuerySet, field: ToManyField, info: GQLInfo) -> QuerySet:
        """
        Paginate a prefetch queryset using a window function partitioned by the given related field.

        Pagination arguments are annotated to the queryset, since they are calculated in the database.
        There is the issue that they might not be available if the queryset is empty after pagination,
        but since they can be different for each prefetch partition, we cannot do anything about that.
        """
        if isinstance(field, ManyToManyField | ManyToManyRel):
            register_for_prefetch_hack(queryset, field)

        related_name = field.remote_field.name

        if self.total_count is not None:
            queryset = add_total_count(queryset, related_name)

        queryset = add_partition_index(queryset, related_name)

        queryset = add_start_index(queryset, self.start)
        queryset = filter_by_start_index(queryset)

        if self.stop is not None:
            queryset = add_stop_index(queryset, self.stop)
            queryset = filter_by_stop_index(queryset)

        return queryset


class Node(InterfaceType):
    """An interface for objects with Global IDs."""

    id = InterfaceField(
        GraphQLNonNull(GraphQLID),
        resolvable_output_type=True,
        description="The Global ID of an object.",
    )


class Connection:
    """A Relay `Connection` for paginating a `QueryType`."""

    def __init__(
        self,
        query_type: type[QueryType],
        /,
        *,
        page_size: int | None = undine_settings.CONNECTION_PAGE_SIZE,
        pagination_handler: type[PaginationHandler] = PaginationHandler,
        description: str | None = None,
    ) -> None:
        """
        Create a new Connection.

        :param query_type: The `QueryType` to use for the Connection.
        :param page_size: Maximum number of items to return in a page. No limit if `None`.
        :param pagination_handler: Handler to use for paginating the Connection.
        :param description: Description for the Connection.
        """
        self.query_type = query_type
        self.page_size = page_size
        self.pagination_handler = pagination_handler
        self.description = description


def add_partition_index(queryset: QuerySet, related_name: str) -> QuerySet:
    """Add an index to each instance in the queryset, partitioned by the given related name."""
    return queryset.alias(
        **{
            undine_settings.CONNECTION_INDEX_KEY: (
                Window(
                    expression=RowNumber(),
                    partition_by=F(related_name),
                    order_by=queryset.query.order_by or copy(queryset.model._meta.ordering) or None,
                )
                - Value(1)  # Start from zero.
            ),
        },
    )


def add_total_count(queryset: QuerySet, related_name: str) -> QuerySet:
    """Add an annotation to the given queryset with the total count of objects in the queryset."""
    total_count = total_count_subquery(queryset, related_name)
    return queryset.annotate(**{undine_settings.CONNECTION_TOTAL_COUNT_KEY: total_count})


def total_count_subquery(queryset: QuerySet, related_name: str) -> SubqueryCount:
    """Get a subquery for calculating total count, partitioned by the given related name."""
    return SubqueryCount(queryset=queryset.filter(**{related_name: OuterRef(related_name)}))


def add_start_index(queryset: QuerySet, start: int | CombinableExpression) -> QuerySet:
    """Add an annotation to the given queryset with the start index of the current page."""
    if isinstance(start, int):
        start = Value(start)
    return queryset.annotate(**{undine_settings.CONNECTION_START_INDEX_KEY: start})


def filter_by_start_index(queryset: QuerySet) -> QuerySet:
    """Filter out all items before the start index of the current page."""
    start = F(undine_settings.CONNECTION_START_INDEX_KEY)
    return queryset.filter(**{f"{undine_settings.CONNECTION_INDEX_KEY}__gte": start})


def add_stop_index(queryset: QuerySet, stop: int | CombinableExpression) -> QuerySet:
    """Add an annotation to the given queryset with the stop index of the current page."""
    if isinstance(stop, int):
        stop = Value(stop)
    return queryset.annotate(**{undine_settings.CONNECTION_STOP_INDEX_KEY: stop})


def filter_by_stop_index(queryset: QuerySet) -> QuerySet:
    """Filter out all items on or after the stop index of the current page."""
    stop = F(undine_settings.CONNECTION_STOP_INDEX_KEY)
    return queryset.filter(**{f"{undine_settings.CONNECTION_INDEX_KEY}__lt": stop})


def encode_base64(string: str) -> str:
    return base64.b64encode(string.encode("utf-8")).decode("ascii")


def decode_base64(string: str) -> str:
    return base64.b64decode(string.encode("ascii")).decode("utf-8")


def offset_to_cursor(typename: str, offset: int) -> str:
    """Create the cursor string from an offset."""
    return encode_base64(f"connection:{typename}:{offset}")


def cursor_to_offset(typename: str, cursor: str) -> int:
    """Extract the offset from the cursor string."""
    return int(decode_base64(cursor).removeprefix(f"connection:{typename}:"))


def to_global_id(typename: str, object_id: str | int) -> str:
    """
    Takes a typename and an object ID specific to that type,
    and returns a "Global ID" that is unique among all types.
    """
    return encode_base64(f"ID:{typename}:{serialize_id(object_id)}")


def from_global_id(global_id: str) -> tuple[str, str | int]:
    """
    Takes the "Global ID" created by `to_global_id`,
    and returns the typename and object ID used to create it.
    """
    global_id = decode_base64(global_id)
    _, typename, object_id = global_id.split(":")
    if object_id.isdigit():
        return typename, int(object_id)
    return typename, object_id


PageInfoType = get_or_create_graphql_object_type(
    name="PageInfo",
    description="Information about the current state of the pagination.",
    fields={
        "hasNextPage": GraphQLField(
            GraphQLNonNull(GraphQLBoolean),
            description="Are there more items after the current page?",
        ),
        "hasPreviousPage": GraphQLField(
            GraphQLNonNull(GraphQLBoolean),
            description="Are there more items before the current page?",
        ),
        "startCursor": GraphQLField(
            GraphQLString,  # null if no results
            description=(
                "Value of the first cursor in the current page. "
                "Use as the value for the `before` argument to paginate backwards."
            ),
        ),
        "endCursor": GraphQLField(
            GraphQLString,  # null if no results
            description=(
                "Value of the last cursor in the current page. "
                "Use as the value for the `after` argument to paginate forwards."
            ),
        ),
    },
)



from __future__ import annotations

import dataclasses
from asyncio import iscoroutinefunction
from typing import TYPE_CHECKING, Any, Generic, Optional, TypeAlias

from asgiref.sync import sync_to_async
from django.db.models import QuerySet
from django.db.models.manager import BaseManager
from graphql import GraphQLID, GraphQLObjectType

from undine import QueryType
from undine.dataclasses import OptimizationWithPagination
from undine.exceptions import (
    GraphQLNodeIDFieldTypeError,
    GraphQLNodeInterfaceMissingError,
    GraphQLNodeInvalidGlobalIDError,
    GraphQLNodeMissingIDFieldError,
    GraphQLNodeObjectTypeMissingError,
    GraphQLNodeQueryTypeMissingError,
    GraphQLNodeTypeNotObjectTypeError,
)
from undine.optimizer.optimizer import optimize_async, optimize_sync
from undine.optimizer.prefetch_hack import evaluate_with_prefetch_hack_async, evaluate_with_prefetch_hack_sync
from undine.relay import Node, from_global_id, offset_to_cursor, to_global_id
from undine.settings import undine_settings
from undine.typing import ConnectionDict, NodeDict, PageInfoDict, TModel
from undine.utils.graphql.undine_extensions import get_undine_query_type
from undine.utils.graphql.utils import get_queried_field_name, get_underlying_type, pre_evaluate_request_user
from undine.utils.reflection import get_root_and_info_params, is_subclass

if TYPE_CHECKING:
    from collections.abc import Callable
    from types import FunctionType

    from django.db.models import Model, Q
    from graphql.pyutils import AwaitableOrValue

    from undine import Entrypoint, Field, InterfaceType, UnionType
    from undine.optimizer.optimizer import QueryOptimizer
    from undine.relay import Connection, PaginationHandler
    from undine.typing import GQLInfo

__all__ = [
    "ConnectionResolver",
    "EntrypointFunctionResolver",
    "FieldFunctionResolver",
    "GlobalIDResolver",
    "InterfaceResolver",
    "ModelAttributeResolver",
    "ModelManyRelatedFieldResolver",
    "ModelSingleRelatedFieldResolver",
    "NamedTupleFieldResolver",
    "NestedConnectionResolver",
    "NestedQueryTypeManyResolver",
    "NestedQueryTypeSingleResolver",
    "NodeResolver",
    "QueryTypeManyResolver",
    "QueryTypeSingleResolver",
    "TypedDictFieldResolver",
    "UnionTypeResolver",
]


@dataclasses.dataclass(frozen=True, slots=True)
class EntrypointFunctionResolver:
    """Resolves an `Entrypoint` using the given function."""

    func: FunctionType | Callable[..., Any]
    entrypoint: Entrypoint

    root_param: str | None = dataclasses.field(default=None, init=False)
    info_param: str | None = dataclasses.field(default=None, init=False)

    def __post_init__(self) -> None:
        params = get_root_and_info_params(self.func)
        object.__setattr__(self, "root_param", params.root_param)
        object.__setattr__(self, "info_param", params.info_param)

    def __call__(self, root: Any, info: GQLInfo, **kwargs: Any) -> Any:
        if undine_settings.ASYNC and iscoroutinefunction(self.func):
            return self.run_async(root, info, **kwargs)
        return self.run_sync(root, info, **kwargs)

    def run_sync(self, root: Any, info: GQLInfo, **kwargs: Any) -> Any:
        self.set_kwargs(kwargs, root, info)
        result = self.func(**kwargs)
        self.check_permissions(root, info, result)
        return result

    async def run_async(self, root: Any, info: GQLInfo, **kwargs: Any) -> Any:
        # Fetch user eagerly so that its available e.g. for permission checks in synchronous parts of the code.
        await pre_evaluate_request_user(info)

        self.set_kwargs(kwargs, root, info)
        result = await self.func(**kwargs)
        self.check_permissions(root, info, result)
        return result

    def set_kwargs(self, kwargs: dict[str, Any], root: Any, info: GQLInfo) -> None:
        if self.root_param is not None:
            kwargs[self.root_param] = root
        if self.info_param is not None:
            kwargs[self.info_param] = info

    def check_permissions(self, root: Any, info: GQLInfo, result: Any) -> None:
        if self.entrypoint.permissions_func is not None:
            if self.entrypoint.many:
                for item in result:
                    self.entrypoint.permissions_func(root, info, item)
            else:
                self.entrypoint.permissions_func(root, info, result)

        elif is_subclass(self.entrypoint.ref, QueryType):
            self.entrypoint.ref.__permissions__(result, info)


@dataclasses.dataclass(frozen=True, slots=True)
class FieldFunctionResolver:
    """Resolves a `Field` using the given function."""

    func: FunctionType | Callable[..., Any]
    field: Field

    root_param: str | None = dataclasses.field(default=None, init=False)
    info_param: str | None = dataclasses.field(default=None, init=False)

    def __post_init__(self) -> None:
        params = get_root_and_info_params(self.func)
        object.__setattr__(self, "root_param", params.root_param)
        object.__setattr__(self, "info_param", params.info_param)

    def __call__(self, root: Any, info: GQLInfo, **kwargs: Any) -> Any:
        if undine_settings.ASYNC and iscoroutinefunction(self.func):
            return self.run_async(root, info, **kwargs)
        return self.run_sync(root, info, **kwargs)

    def run_sync(self, root: Any, info: GQLInfo, **kwargs: Any) -> Any:
        self.set_kwargs(kwargs, root, info)
        result = self.func(**kwargs)
        self.check_permissions(root, info, result)
        return result

    async def run_async(self, root: Any, info: GQLInfo, **kwargs: Any) -> Any:
        self.set_kwargs(kwargs, root, info)
        result = await self.func(**kwargs)
        self.check_permissions(root, info, result)
        return result

    def set_kwargs(self, kwargs: dict[str, Any], root: Any, info: GQLInfo) -> None:
        if self.root_param is not None:
            kwargs[self.root_param] = root
        if self.info_param is not None:
            kwargs[self.info_param] = info

    def check_permissions(self, root: Any, info: GQLInfo, result: Any) -> None:
        if self.field.permissions_func is not None:
            if self.field.many:
                for item in result:
                    self.field.permissions_func(root, info, item)
            else:
                self.field.permissions_func(root, info, result)

        elif is_subclass(self.field.ref, QueryType):
            self.field.ref.__permissions__(result, info)


# Model field resolvers


@dataclasses.dataclass(frozen=True, slots=True)
class ModelAttributeResolver:
    """Resolves a model field or annotation to a value by attribute access."""

    field: Field

    static: bool = True
    """
    If the attribute is queried multiple times in the same operation, should it return
    different values, for example, based on input arguments?
    """

    def __call__(self, root: Model, info: GQLInfo, **kwargs: Any) -> Any:
        field_name = self.field.field_name
        if not self.static:
            field_name = get_queried_field_name(field_name, info)

        value = getattr(root, field_name, None)

        if self.field.permissions_func is not None:
            self.field.permissions_func(root, info, value)

        return value


@dataclasses.dataclass(frozen=True, slots=True)
class ModelSingleRelatedFieldResolver(Generic[TModel]):
    """Resolves single-related model field to its primary key."""

    field: Field

    def __call__(self, root: Model, info: GQLInfo, **kwargs: Any) -> int | None:
        value: TModel | None = getattr(root, self.field.field_name, None)

        if value is None:
            return None

        if self.field.permissions_func is not None:
            self.field.permissions_func(root, info, value)

        return value.pk


@dataclasses.dataclass(frozen=True, slots=True)
class ModelManyRelatedFieldResolver(Generic[TModel]):
    """Resolves a many-related model field to a list of their primary keys."""

    field: Field

    def __call__(self, root: Model, info: GQLInfo, **kwargs: Any) -> list[TModel]:
        field_name = get_queried_field_name(self.field.field_name, info)
        manager: BaseManager[TModel] = getattr(root, field_name)
        instances = list(manager.get_queryset())

        if self.field.permissions_func is not None:
            for instance in instances:
                self.field.permissions_func(root, info, instance)

        return [instance.pk for instance in instances]


@dataclasses.dataclass(frozen=True, slots=True)
class ModelGenericForeignKeyResolver(Generic[TModel]):
    """Resolves generic foreign key field to its related model instance."""

    field: Field

    def __call__(self, root: Model, info: GQLInfo, **kwargs: Any) -> TModel | None:
        value: TModel | None = getattr(root, self.field.field_name, None)

        if value is None:
            return None

        if self.field.permissions_func is not None:
            self.field.permissions_func(root, info, value)

        return value


# Query type resolvers


@dataclasses.dataclass(frozen=True, slots=True)
class QueryTypeSingleResolver(Generic[TModel]):
    """Top-level resolver for fetching a single model object via a QueryType."""

    query_type: type[QueryType[TModel]]
    entrypoint: Entrypoint

    def __call__(self, root: Any, info: GQLInfo, **kwargs: Any) -> AwaitableOrValue[TModel | None]:
        if undine_settings.ASYNC:
            return self.run_async(root, info, **kwargs)
        return self.run_sync(root, info, **kwargs)

    def run_sync(self, root: Any, info: GQLInfo, **kwargs: Any) -> TModel | None:
        queryset = self.query_type.__get_queryset__(info)
        instance = optimize_sync(queryset, info, **kwargs)

        if instance is not None:
            self.check_permissions(root, info, instance)
        return instance

    async def run_async(self, root: Any, info: GQLInfo, **kwargs: Any) -> TModel | None:
        # Fetch user eagerly so that its available e.g. for permission checks in synchronous parts of the code.
        await pre_evaluate_request_user(info)

        queryset = self.query_type.__get_queryset__(info)
        instance = await optimize_async(queryset, info, **kwargs)

        if instance is not None:
            self.check_permissions(root, info, instance)
        return instance

    def check_permissions(self, root: Any, info: GQLInfo, instance: TModel) -> None:
        if self.entrypoint.permissions_func is not None:
            self.entrypoint.permissions_func(root, info, instance)
        else:
            self.query_type.__permissions__(instance, info)


@dataclasses.dataclass(frozen=True, slots=True)
class QueryTypeManyResolver(Generic[TModel]):
    """Top-level resolver for fetching a set of model objects via a QueryType."""

    query_type: type[QueryType[TModel]]
    entrypoint: Entrypoint

    additional_filter: Optional[Q] = None  # noqa: UP045

    def __call__(self, root: Any, info: GQLInfo, **kwargs: Any) -> AwaitableOrValue[list[TModel]]:
        if undine_settings.ASYNC:
            return self.run_async(root, info, **kwargs)
        return self.run_sync(root, info, **kwargs)

    def run_sync(self, root: Any, info: GQLInfo, **kwargs: Any) -> list[TModel]:
        queryset = self.get_queryset(info)
        instances = optimize_sync(queryset, info)
        self.check_permissions(root, info, instances)
        return instances

    async def run_async(self, root: Any, info: GQLInfo, **kwargs: Any) -> list[TModel]:
        # Fetch user eagerly so that its available e.g. for permission checks in synchronous parts of the code.
        await pre_evaluate_request_user(info)

        queryset = self.get_queryset(info)
        instances = await optimize_async(queryset, info)
        self.check_permissions(root, info, instances)
        return instances

    def get_queryset(self, info: GQLInfo) -> QuerySet[TModel]:
        queryset = self.query_type.__get_queryset__(info)
        if self.additional_filter is not None:
            queryset = queryset.filter(self.additional_filter)
        return queryset

    def check_permissions(self, root: Any, info: GQLInfo, instances: list[TModel]) -> None:
        for instance in instances:
            if self.entrypoint.permissions_func is not None:
                self.entrypoint.permissions_func(root, info, instance)
            else:
                self.query_type.__permissions__(instance, info)


@dataclasses.dataclass(frozen=True, slots=True)
class NestedQueryTypeSingleResolver(Generic[TModel]):
    """Resolves a single-related field pointing to another QueryType."""

    query_type: type[QueryType[TModel]]
    field: Field

    def __call__(self, root: Model, info: GQLInfo, **kwargs: Any) -> TModel | None:
        instance: TModel | None = getattr(root, self.field.field_name, None)

        if instance is not None:
            if self.field.permissions_func is not None:
                self.field.permissions_func(root, info, instance)
            else:
                self.query_type.__permissions__(instance, info)

        return instance


@dataclasses.dataclass(frozen=True, slots=True)
class NestedQueryTypeManyResolver(Generic[TModel]):
    """Resolves a many-related field pointing to another QueryType."""

    query_type: type[QueryType[TModel]]
    field: Field

    def __call__(self, root: Model, info: GQLInfo, **kwargs: Any) -> list[TModel]:
        field_name = get_queried_field_name(self.field.field_name, info)

        instances: list[TModel] = getattr(root, field_name)
        if isinstance(instances, BaseManager):
            instances = list(instances.get_queryset())

        for instance in instances:
            if self.field.permissions_func is not None:
                self.field.permissions_func(root, info, instance)
            else:
                self.query_type.__permissions__(instance, info)

        return instances


# Relay


@dataclasses.dataclass(frozen=True, slots=True)
class GlobalIDResolver:
    """Resolves a model primary key as a Global ID."""

    typename: str

    def __call__(self, root: Model, info: GQLInfo, **kwargs: Any) -> str:
        return to_global_id(self.typename, root.pk)


@dataclasses.dataclass(frozen=True, slots=True)
class NodeResolver(Generic[TModel]):
    """Resolves a model instance through a Global ID."""

    entrypoint: Entrypoint

    def __call__(self, root: Any, info: GQLInfo, **kwargs: Any) -> AwaitableOrValue[TModel | None]:
        try:
            typename, object_id = from_global_id(kwargs["id"])
        except Exception as error:
            raise GraphQLNodeInvalidGlobalIDError(value=kwargs["id"]) from error

        object_type = info.schema.get_type(typename)
        if object_type is None:
            raise GraphQLNodeObjectTypeMissingError(typename=typename)

        if not isinstance(object_type, GraphQLObjectType):
            raise GraphQLNodeTypeNotObjectTypeError(typename=typename)

        query_type = get_undine_query_type(object_type)
        if query_type is None:
            raise GraphQLNodeQueryTypeMissingError(typename=typename)

        if Node not in query_type.__interfaces__:
            raise GraphQLNodeInterfaceMissingError(typename=typename)

        field: Field | None = query_type.__field_map__.get("id")
        if field is None:
            raise GraphQLNodeMissingIDFieldError(typename=typename)

        field_type = get_underlying_type(field.get_field_type())  # type: ignore[type-var]
        if field_type is not GraphQLID:
            raise GraphQLNodeIDFieldTypeError(typename=typename)

        resolver = QueryTypeSingleResolver(query_type=query_type, entrypoint=self.entrypoint)
        return resolver(root, info, pk=object_id)


@dataclasses.dataclass(frozen=True, slots=True)
class ConnectionResolver(Generic[TModel]):
    """Resolves a connection of items."""

    connection: Connection
    entrypoint: Entrypoint

    def __call__(self, root: Any, info: GQLInfo, **kwargs: Any) -> AwaitableOrValue[ConnectionDict[TModel]]:
        if undine_settings.ASYNC:
            return self.run_async(root, info)
        return self.run_sync(root, info)

    def run_sync(self, root: Any, info: GQLInfo) -> ConnectionDict[TModel]:
        results = self.run_optimizer(info)
        instances = evaluate_with_prefetch_hack_sync(results.queryset)
        self.check_permissions(root, info, instances)
        return self.to_connection(instances, pagination=results.pagination)

    async def run_async(self, root: Any, info: GQLInfo) -> ConnectionDict[TModel]:
        # Fetch user eagerly so that its available e.g. for permission checks in synchronous parts of the code.
        await pre_evaluate_request_user(info)

        results = await self.run_optimizer_async(info)
        instances = await evaluate_with_prefetch_hack_async(results.queryset)
        self.check_permissions(root, info, instances)
        return self.to_connection(instances, pagination=results.pagination)

    def get_queryset(self, info: GQLInfo) -> QuerySet[TModel]:
        return self.connection.query_type.__get_queryset__(info)

    def run_optimizer(self, info: GQLInfo) -> OptimizationWithPagination[TModel]:
        queryset = self.get_queryset(info)
        optimizer: QueryOptimizer = undine_settings.OPTIMIZER_CLASS(model=queryset.model, info=info)
        optimizations = optimizer.compile()
        optimized_queryset = optimizations.apply(queryset, info)
        return OptimizationWithPagination(
            queryset=optimized_queryset,
            pagination=optimizations.pagination,  # type: ignore[arg-type]
        )

    async def run_optimizer_async(self, info: GQLInfo) -> OptimizationWithPagination[TModel]:
        queryset = self.get_queryset(info)
        optimizer: QueryOptimizer = undine_settings.OPTIMIZER_CLASS(model=queryset.model, info=info)
        optimizations = optimizer.compile()
        # Applying may call 'queryset.count()'.
        optimized_queryset = await sync_to_async(optimizations.apply)(queryset, info)
        return OptimizationWithPagination(
            queryset=optimized_queryset,
            pagination=optimizations.pagination,  # type: ignore[arg-type]
        )

    def check_permissions(self, root: Any, info: GQLInfo, instances: list[TModel]) -> None:
        for instance in instances:
            if self.entrypoint.permissions_func is not None:
                self.entrypoint.permissions_func(root, info, instance)
            else:
                self.connection.query_type.__permissions__(instance, info)

    def to_connection(self, instances: list[TModel], pagination: PaginationHandler) -> ConnectionDict[TModel]:
        typename = self.connection.query_type.__schema_name__
        edges = [
            NodeDict(
                cursor=offset_to_cursor(typename, pagination.start + index),
                node=instance,
            )
            for index, instance in enumerate(instances)
        ]
        return ConnectionDict(
            totalCount=pagination.total_count or 0,
            pageInfo=PageInfoDict(
                hasNextPage=(
                    False
                    if pagination.stop is None
                    else True
                    if pagination.total_count is None
                    else pagination.stop < pagination.total_count
                ),
                hasPreviousPage=pagination.start > 0,
                startCursor=None if not edges else edges[0]["cursor"],
                endCursor=None if not edges else edges[-1]["cursor"],
            ),
            edges=edges,
        )


@dataclasses.dataclass(frozen=True, slots=True)
class NestedConnectionResolver(Generic[TModel]):
    """Resolves a nested connection from the given field."""

    connection: Connection
    field: Field

    def __call__(self, root: Model, info: GQLInfo, **kwargs: Any) -> ConnectionDict[TModel]:
        field_name = get_queried_field_name(self.field.field_name, info)
        instances = self.get_instances(root, field_name)
        self.check_permissions(root, info, instances)
        return self.to_connection(instances)

    def get_instances(self, root: Model, field_name: str) -> list[TModel]:
        instances: list[TModel] = getattr(root, field_name)
        if isinstance(instances, BaseManager):
            instances = list(instances.get_queryset())
        return instances

    def check_permissions(self, root: Any, info: GQLInfo, instances: list[TModel]) -> None:
        for instance in instances:
            if self.field.permissions_func is not None:
                self.field.permissions_func(root, info, instance)
            else:
                self.connection.query_type.__permissions__(instance, info)

    def get_pagination_params(self, instances: list[TModel]) -> tuple[int, int | None, int | None]:
        total_count: int | None = None
        start: int = 0
        stop: int | None = None

        # Not optimal, as we don't know the actual pagination params if there are no results.
        if instances:
            total_count = getattr(instances[0], undine_settings.CONNECTION_TOTAL_COUNT_KEY, None)
            start = getattr(instances[0], undine_settings.CONNECTION_START_INDEX_KEY, 0)
            stop = getattr(instances[0], undine_settings.CONNECTION_STOP_INDEX_KEY, None)

        return start, stop, total_count

    def to_connection(self, instances: list[TModel]) -> ConnectionDict[TModel]:
        start, stop, total_count = self.get_pagination_params(instances)

        typename = self.connection.query_type.__schema_name__
        edges = [
            NodeDict(
                cursor=offset_to_cursor(typename, start + index),
                node=instance,
            )
            for index, instance in enumerate(instances)
        ]
        return ConnectionDict(
            totalCount=total_count or 0,
            pageInfo=PageInfoDict(
                hasNextPage=(False if stop is None else True if total_count is None else stop < total_count),
                hasPreviousPage=start > 0,
                startCursor=None if not edges else edges[0]["cursor"],
                endCursor=None if not edges else edges[-1]["cursor"],
            ),
            edges=edges,
        )


# UnionType


QuerySetMap: TypeAlias = dict[type[QueryType[TModel]], QuerySet[TModel]]


@dataclasses.dataclass(frozen=True, slots=True)
class UnionTypeResolver(Generic[TModel]):
    """Top-level resolver for fetching a set of model objects from a `UnionType`."""

    union_type: type[UnionType]
    entrypoint: Entrypoint

    def __call__(self, root: Any, info: GQLInfo, **kwargs: Any) -> AwaitableOrValue[list[TModel]]:
        if undine_settings.ASYNC:
            return self.run_async(root, info, **kwargs)
        return self.run_sync(root, info, **kwargs)

    def run_sync(self, root: Any, info: GQLInfo, **kwargs: Any) -> list[TModel]:
        queryset_map = self.optimize(info, **kwargs)
        all_instances = self.fetch_instances(root, info, queryset_map)
        return self.union_type.__process_results__(all_instances, info)

    async def run_async(self, root: Any, info: GQLInfo, **kwargs: Any) -> list[TModel]:
        queryset_map = self.optimize(info, **kwargs)
        all_instances = await self.fetch_instances_async(root, info, queryset_map)
        return self.union_type.__process_results__(all_instances, info)

    def fetch_instances(self, root: Any, info: GQLInfo, queryset_map: QuerySetMap[TModel]) -> list[TModel]:
        all_instances: list[TModel] = []

        for query_type, queryset in queryset_map.items():
            instances = evaluate_with_prefetch_hack_sync(queryset)
            self.check_permissions(root, info, query_type, instances)
            all_instances.extend(instances)

        return all_instances

    async def fetch_instances_async(self, root: Any, info: GQLInfo, queryset_map: QuerySetMap[TModel]) -> list[TModel]:
        all_instances: list[TModel] = []

        for query_type, queryset in queryset_map.items():
            instances = await evaluate_with_prefetch_hack_async(queryset)
            self.check_permissions(root, info, query_type, instances)
            all_instances.extend(instances)

        return all_instances

    def check_permissions(
        self,
        root: Any,
        info: GQLInfo,
        query_type: type[QueryType[TModel]],
        instances: list[TModel],
    ) -> None:
        for instance in instances:
            if self.entrypoint.permissions_func is not None:
                self.entrypoint.permissions_func(root, info, instance)
            else:
                query_type.__permissions__(instance, info)

    def optimize(self, info: GQLInfo, **kwargs: Any) -> QuerySetMap[TModel]:
        queryset_map: QuerySetMap[TModel] = {}

        for model, query_type in self.union_type.__query_types_by_model__.items():
            optimizer: QueryOptimizer = undine_settings.OPTIMIZER_CLASS(model=model, info=info)
            optimizations = optimizer.compile()

            # If nothing is selected from this union type member, don't fetch anything from it
            if not optimizations:
                continue

            args: dict[str, Any] = {}
            filter_key = f"{undine_settings.QUERY_TYPE_FILTER_INPUT_KEY}{model.__name__}"
            order_by_key = f"{undine_settings.QUERY_TYPE_ORDER_INPUT_KEY}{model.__name__}"

            if filter_key in kwargs:
                args[undine_settings.QUERY_TYPE_FILTER_INPUT_KEY] = kwargs[filter_key]
            if order_by_key in kwargs:
                args[undine_settings.QUERY_TYPE_ORDER_INPUT_KEY] = kwargs[order_by_key]

            optimizer.handle_undine_query_type(query_type, args)

            queryset = query_type.__get_queryset__(info)
            optimized_queryset = optimizations.apply(queryset, info)

            queryset_map[query_type] = optimized_queryset[: self.entrypoint.limit]

        return queryset_map


@dataclasses.dataclass(frozen=True, slots=True)
class _UnionTypeConnectionResolver:  # TODO: Implement
    """Resolves a connection of union type items."""

    connection: Connection
    entrypoint: Entrypoint

    def __call__(self, root: Any, info: GQLInfo, **kwargs: Any) -> ConnectionDict[Model]:
        msg = "Union connections are not implemented yet."
        raise NotImplementedError(msg)


# InterfaceType


@dataclasses.dataclass(frozen=True, slots=True)
class InterfaceResolver(Generic[TModel]):
    """Resolves an interface type to all of its implementations."""

    interface: type[InterfaceType]
    entrypoint: Entrypoint

    def __call__(self, root: Any, info: GQLInfo, **kwargs: Any) -> AwaitableOrValue[list[TModel]]:
        if undine_settings.ASYNC:
            return self.run_sync_async(root, info, **kwargs)
        return self.run_sync(root, info, **kwargs)

    def run_sync(self, root: Any, info: GQLInfo, **kwargs: Any) -> list[TModel]:
        queryset_map = self.optimize(info, **kwargs)
        all_instances = self.fetch_instances(info, root, queryset_map)
        return self.interface.__process_results__(all_instances, info)

    async def run_sync_async(self, root: Any, info: GQLInfo, **kwargs: Any) -> list[TModel]:
        queryset_map = self.optimize(info, **kwargs)
        all_instances = await self.fetch_instances_async(info, root, queryset_map)
        return self.interface.__process_results__(all_instances, info)

    def fetch_instances(self, info: GQLInfo, root: Any, queryset_map: QuerySetMap[TModel]) -> list[TModel]:
        all_instances: list[TModel] = []

        for query_type, queryset in queryset_map.items():
            instances = evaluate_with_prefetch_hack_sync(queryset)
            self.check_permissions(info, root, query_type, instances)
            all_instances.extend(instances)

        return all_instances

    async def fetch_instances_async(self, info: GQLInfo, root: Any, queryset_map: QuerySetMap[TModel]) -> list[TModel]:
        all_instances: list[TModel] = []

        for query_type, queryset in queryset_map.items():
            instances = await evaluate_with_prefetch_hack_async(queryset)
            self.check_permissions(info, root, query_type, instances)
            all_instances.extend(instances)

        return all_instances

    def check_permissions(
        self,
        info: GQLInfo,
        root: Any,
        query_type: type[QueryType[TModel]],
        instances: list[TModel],
    ) -> None:
        for instance in instances:
            if self.entrypoint.permissions_func is not None:
                self.entrypoint.permissions_func(root, info, instance)
            else:
                query_type.__permissions__(instance, info)

    def optimize(self, info: GQLInfo, **kwargs: Any) -> QuerySetMap[TModel]:
        queryset_map: QuerySetMap[TModel] = {}

        for query_type in self.interface.__concrete_implementations__():
            model = query_type.__model__
            optimizer: QueryOptimizer = undine_settings.OPTIMIZER_CLASS(model=model, info=info)
            optimizations = optimizer.compile()

            # If nothing is selected from this interface implementation, don't fetch anything from it
            if not optimizations:
                continue

            args: dict[str, Any] = {}
            filter_key = f"{undine_settings.QUERY_TYPE_FILTER_INPUT_KEY}{model.__name__}"
            order_by_key = f"{undine_settings.QUERY_TYPE_ORDER_INPUT_KEY}{model.__name__}"

            if filter_key in kwargs:
                args[undine_settings.QUERY_TYPE_FILTER_INPUT_KEY] = kwargs[filter_key]
            if order_by_key in kwargs:
                args[undine_settings.QUERY_TYPE_ORDER_INPUT_KEY] = kwargs[order_by_key]

            optimizer.handle_undine_query_type(query_type, args)

            queryset = query_type.__get_queryset__(info)
            optimized_queryset = optimizations.apply(queryset, info)

            queryset_map[query_type] = optimized_queryset[: self.entrypoint.limit]

        return queryset_map


@dataclasses.dataclass(frozen=True, slots=True)
class _InterfaceConnectionResolver:  # TODO: Implement
    """Resolves a connection of interface type items."""

    connection: Connection
    entrypoint: Entrypoint

    def __call__(self, root: Any, info: GQLInfo, **kwargs: Any) -> ConnectionDict[Model]:
        msg = "Interface connections are not implemented yet."
        raise NotImplementedError(msg)


# Miscellaneous


@dataclasses.dataclass(frozen=True, slots=True)
class TypedDictFieldResolver:
    """Resolves a typed dict field."""

    key: str
    """
    The actual key in the typed dict,
    which may have been converted to camel case for the GraphQL ObjectType.
    """

    def __call__(self, root: dict[str, Any], info: GQLInfo, **kwargs: Any) -> Any:
        return root.get(self.key, None)


@dataclasses.dataclass(frozen=True, slots=True)
class NamedTupleFieldResolver:
    """Resolves a named tuple field."""

    attr: str
    """
    The actual attribute in the named tuple,
    which may have been converted to camel case for the GraphQL ObjectType.
    """

    def __call__(self, root: object, info: GQLInfo, **kwargs: Any) -> Any:
        return getattr(root, self.attr, None)



from __future__ import annotations

import dataclasses
from types import FunctionType, SimpleNamespace
from typing import TYPE_CHECKING, Any, Generic

from asgiref.sync import iscoroutinefunction, sync_to_async
from django.db.models import Q
from graphql import GraphQLError, Undefined

from undine import MutationType
from undine.exceptions import GraphQLErrorGroup, GraphQLMissingLookupFieldError, GraphQLModelConstraintViolationError
from undine.parsers import parse_model_relation_info
from undine.settings import undine_settings
from undine.typing import MutationKind, TModel
from undine.utils.graphql.utils import graphql_error_path, pre_evaluate_request_user
from undine.utils.model_utils import convert_integrity_errors, get_default_manager
from undine.utils.mutation_data import MutationData, MutationManyData, get_mutation_data, get_related_info
from undine.utils.mutation_tree import bulk_mutate, mutate
from undine.utils.reflection import is_subclass

from .query import QueryTypeManyResolver, QueryTypeSingleResolver

if TYPE_CHECKING:
    from django.db.models import Model
    from graphql.pyutils import AwaitableOrValue

    from undine import Entrypoint, GQLInfo, Input, QueryType
    from undine.typing import EntrypointPermFunc, InputPermFunc, ValidatorFunc

__all__ = [
    "BulkCreateResolver",
    "BulkDeleteResolver",
    "BulkUpdateResolver",
    "CreateResolver",
    "CustomResolver",
    "DeleteResolver",
    "UpdateResolver",
]


@dataclasses.dataclass(frozen=True, slots=True)
class CreateResolver(Generic[TModel]):
    """Resolves a mutation for creating a model instance using."""

    mutation_type: type[MutationType[TModel]]
    entrypoint: Entrypoint

    def __call__(self, root: Any, info: GQLInfo, **kwargs: Any) -> AwaitableOrValue[TModel | None]:
        if undine_settings.ASYNC:
            return self.run_async(root, info, **kwargs)
        return self.run_sync(root, info, **kwargs)

    @property
    def model(self) -> type[TModel]:
        return self.mutation_type.__model__  # type: ignore[return-value]

    @property
    def query_type(self) -> type[QueryType[TModel]]:
        return self.mutation_type.__query_type__()

    def run_sync(self, root: Any, info: GQLInfo, **kwargs: Any) -> TModel | None:
        data: dict[str, Any] = kwargs[undine_settings.MUTATION_INPUT_DATA_KEY]

        mutation_data = get_mutation_data(
            model=self.model,
            data=data,
            mutation_type=self.mutation_type,
        )

        _pre_mutation_chain(
            root=root,
            info=info,
            mutation_data=mutation_data,
            mutation_type=self.mutation_type,
            permissions_func=self.entrypoint.permissions_func,
        )

        previous_data = mutation_data.previous_data

        instance = mutate(mutation_data, model=self.model)

        self.mutation_type.__after__(instance=instance, info=info, previous_data=previous_data)

        resolver = QueryTypeSingleResolver(query_type=self.query_type, entrypoint=self.entrypoint)
        return resolver.run_sync(root, info, pk=instance.pk)

    async def run_async(self, root: Any, info: GQLInfo, **kwargs: Any) -> TModel | None:
        # Fetch user eagerly so that its available e.g. for permission checks in synchronous parts of the code.
        await pre_evaluate_request_user(info)

        data: dict[str, Any] = kwargs[undine_settings.MUTATION_INPUT_DATA_KEY]

        mutation_data: MutationData = await sync_to_async(get_mutation_data)(
            model=self.model,
            data=data,
            mutation_type=self.mutation_type,
        )

        _pre_mutation_chain(
            root=root,
            info=info,
            mutation_data=mutation_data,
            mutation_type=self.mutation_type,
            permissions_func=self.entrypoint.permissions_func,
        )

        previous_data = mutation_data.previous_data

        instance: TModel = await sync_to_async(mutate)(mutation_data, model=self.model)

        self.mutation_type.__after__(instance=instance, info=info, previous_data=previous_data)

        resolver = QueryTypeSingleResolver(query_type=self.query_type, entrypoint=self.entrypoint)
        return await resolver.run_async(root, info, pk=instance.pk)


@dataclasses.dataclass(frozen=True, slots=True)
class UpdateResolver(Generic[TModel]):
    """Resolves a mutation for updating a model instance."""

    mutation_type: type[MutationType[TModel]]
    entrypoint: Entrypoint

    def __call__(self, root: Any, info: GQLInfo, **kwargs: Any) -> AwaitableOrValue[TModel | None]:
        if undine_settings.ASYNC:
            return self.run_async(root, info, **kwargs)
        return self.run_sync(root, info, **kwargs)

    @property
    def model(self) -> type[TModel]:
        return self.mutation_type.__model__  # type: ignore[return-value]

    @property
    def query_type(self) -> type[QueryType[TModel]]:
        return self.mutation_type.__query_type__()

    def run_sync(self, root: Any, info: GQLInfo, **kwargs: Any) -> TModel | None:
        data: dict[str, Any] = kwargs[undine_settings.MUTATION_INPUT_DATA_KEY]

        if "pk" not in data:
            raise GraphQLMissingLookupFieldError(model=self.model, key="pk")

        mutation_data = get_mutation_data(
            model=self.model,
            data=data,
            mutation_type=self.mutation_type,
        )

        _pre_mutation_chain(
            root=root,
            info=info,
            mutation_data=mutation_data,
            mutation_type=self.mutation_type,
            permissions_func=self.entrypoint.permissions_func,
        )

        previous_data = mutation_data.previous_data

        instance = mutate(mutation_data, model=self.model)

        self.mutation_type.__after__(instance=instance, info=info, previous_data=previous_data)

        resolver = QueryTypeSingleResolver(query_type=self.query_type, entrypoint=self.entrypoint)
        return resolver.run_sync(root, info, pk=instance.pk)

    async def run_async(self, root: Any, info: GQLInfo, **kwargs: Any) -> TModel | None:
        # Fetch user eagerly so that its available e.g. for permission checks in synchronous parts of the code.
        await pre_evaluate_request_user(info)

        data: dict[str, Any] = kwargs[undine_settings.MUTATION_INPUT_DATA_KEY]

        if "pk" not in data:
            raise GraphQLMissingLookupFieldError(model=self.model, key="pk")

        mutation_data: MutationData = await sync_to_async(get_mutation_data)(
            model=self.model,
            data=data,
            mutation_type=self.mutation_type,
        )

        _pre_mutation_chain(
            root=root,
            info=info,
            mutation_data=mutation_data,
            mutation_type=self.mutation_type,
            permissions_func=self.entrypoint.permissions_func,
        )

        previous_data = mutation_data.previous_data

        instance: TModel = await sync_to_async(mutate)(mutation_data, model=self.model)

        self.mutation_type.__after__(instance=instance, info=info, previous_data=previous_data)

        resolver = QueryTypeSingleResolver(query_type=self.query_type, entrypoint=self.entrypoint)
        return await resolver.run_async(root, info, pk=instance.pk)


@dataclasses.dataclass(frozen=True, slots=True)
class DeleteResolver(Generic[TModel]):
    """Resolves a mutation for deleting a model instance."""

    mutation_type: type[MutationType]
    entrypoint: Entrypoint

    def __call__(self, root: Any, info: GQLInfo, **kwargs: Any) -> AwaitableOrValue[SimpleNamespace]:
        if undine_settings.ASYNC:
            return self.run_async(root, info, **kwargs)
        return self.run_sync(root, info, **kwargs)

    @property
    def model(self) -> type[TModel]:
        return self.mutation_type.__model__  # type: ignore[return-value]

    def run_sync(self, root: Any, info: GQLInfo, **kwargs: Any) -> SimpleNamespace:
        data: dict[str, Any] = kwargs[undine_settings.MUTATION_INPUT_DATA_KEY]

        pk: Any = data.get("pk", Undefined)
        if pk is Undefined:
            raise GraphQLMissingLookupFieldError(model=self.model, key="pk")

        mutation_data = get_mutation_data(
            model=self.model,
            data=data,
            mutation_type=self.mutation_type,
        )

        _pre_mutation_chain(
            root=root,
            info=info,
            mutation_data=mutation_data,
            mutation_type=self.mutation_type,
            permissions_func=self.entrypoint.permissions_func,
        )

        previous_data = mutation_data.previous_data
        instance = mutation_data.instance

        with convert_integrity_errors(GraphQLModelConstraintViolationError):
            instance.delete()

        self.mutation_type.__after__(instance=instance, info=info, previous_data=previous_data)

        return SimpleNamespace(pk=pk)

    async def run_async(self, root: Any, info: GQLInfo, **kwargs: Any) -> SimpleNamespace:
        # Fetch user eagerly so that its available e.g. for permission checks in synchronous parts of the code.
        await pre_evaluate_request_user(info)

        data: dict[str, Any] = kwargs[undine_settings.MUTATION_INPUT_DATA_KEY]

        pk: Any = data.get("pk", Undefined)
        if pk is Undefined:
            raise GraphQLMissingLookupFieldError(model=self.model, key="pk")

        mutation_data: MutationData = await sync_to_async(get_mutation_data)(
            model=self.model,
            data=data,
            mutation_type=self.mutation_type,
        )

        _pre_mutation_chain(
            root=root,
            info=info,
            mutation_data=mutation_data,
            mutation_type=self.mutation_type,
            permissions_func=self.entrypoint.permissions_func,
        )

        previous_data = mutation_data.previous_data
        instance = mutation_data.instance

        with convert_integrity_errors(GraphQLModelConstraintViolationError):
            await instance.adelete()

        self.mutation_type.__after__(instance=instance, info=info, previous_data=previous_data)

        return SimpleNamespace(pk=pk)


# Bulk


@dataclasses.dataclass(frozen=True, slots=True)
class BulkCreateResolver(Generic[TModel]):
    """Resolves a bulk create mutation for creating a list of model instances."""

    mutation_type: type[MutationType[TModel]]
    entrypoint: Entrypoint

    def __call__(self, root: Any, info: GQLInfo, **kwargs: Any) -> AwaitableOrValue[list[TModel]]:
        if undine_settings.ASYNC:
            return self.run_async(root, info, **kwargs)
        return self.run_sync(root, info, **kwargs)

    @property
    def model(self) -> type[TModel]:
        return self.mutation_type.__model__  # type: ignore[return-value]

    @property
    def query_type(self) -> type[QueryType[TModel]]:
        return self.mutation_type.__query_type__()

    def run_sync(self, root: Any, info: GQLInfo, **kwargs: Any) -> list[TModel]:
        data: list[dict[str, Any]] = kwargs[undine_settings.MUTATION_INPUT_DATA_KEY]

        mutation_datas = get_mutation_data(
            model=self.model,
            data=data,
            mutation_type=self.mutation_type,
        )

        _pre_mutation_chain_many(
            root=root,
            info=info,
            mutation_datas=mutation_datas,
            mutation_type=self.mutation_type,
            permissions_func=self.entrypoint.permissions_func,
        )

        previous_datas = [mutation_data.previous_data for mutation_data in mutation_datas]

        instances = bulk_mutate(mutation_datas, model=self.model)

        for instance, pre_data in zip(instances, previous_datas, strict=True):
            self.mutation_type.__after__(instance=instance, info=info, previous_data=pre_data)

        resolver = QueryTypeManyResolver(
            query_type=self.query_type,
            entrypoint=self.entrypoint,
            additional_filter=Q(pk__in=[instance.pk for instance in instances]),
        )
        return resolver.run_sync(root, info)

    async def run_async(self, root: Any, info: GQLInfo, **kwargs: Any) -> list[TModel]:
        # Fetch user eagerly so that its available e.g. for permission checks in synchronous parts of the code.
        await pre_evaluate_request_user(info)

        data: list[dict[str, Any]] = kwargs[undine_settings.MUTATION_INPUT_DATA_KEY]

        mutation_datas: list[MutationData] = await sync_to_async(get_mutation_data)(  # type: ignore[assignment]
            model=self.model,
            data=data,  # type: ignore[arg-type]
            mutation_type=self.mutation_type,
        )

        _pre_mutation_chain_many(
            root=root,
            info=info,
            mutation_datas=mutation_datas,
            mutation_type=self.mutation_type,
            permissions_func=self.entrypoint.permissions_func,
        )

        previous_datas = [mutation_data.previous_data for mutation_data in mutation_datas]

        instances: list[TModel] = await sync_to_async(bulk_mutate)(mutation_datas, model=self.model)  # type: ignore[arg-type,assignment]

        for instance, pre_data in zip(instances, previous_datas, strict=True):
            self.mutation_type.__after__(instance=instance, info=info, previous_data=pre_data)

        resolver = QueryTypeManyResolver(
            query_type=self.query_type,
            entrypoint=self.entrypoint,
            additional_filter=Q(pk__in=[instance.pk for instance in instances]),
        )
        return await resolver.run_async(root, info)


@dataclasses.dataclass(frozen=True, slots=True)
class BulkUpdateResolver(Generic[TModel]):
    """Resolves a bulk update mutation for updating a list of model instances."""

    mutation_type: type[MutationType[TModel]]
    entrypoint: Entrypoint

    def __call__(self, root: Any, info: GQLInfo, **kwargs: Any) -> AwaitableOrValue[list[TModel]]:
        if undine_settings.ASYNC:
            return self.run_async(root, info, **kwargs)
        return self.run_sync(root, info, **kwargs)

    @property
    def model(self) -> type[TModel]:
        return self.mutation_type.__model__  # type: ignore[return-value]

    @property
    def query_type(self) -> type[QueryType[TModel]]:
        return self.mutation_type.__query_type__()

    def run_sync(self, root: Any, info: GQLInfo, **kwargs: Any) -> list[TModel]:
        data: list[dict[str, Any]] = kwargs[undine_settings.MUTATION_INPUT_DATA_KEY]

        mutation_datas = get_mutation_data(
            model=self.model,
            data=data,
            mutation_type=self.mutation_type,
        )

        _pre_mutation_chain_many(
            root=root,
            info=info,
            mutation_datas=mutation_datas,
            mutation_type=self.mutation_type,
            permissions_func=self.entrypoint.permissions_func,
        )

        previous_datas = [mutation_data.previous_data for mutation_data in mutation_datas]

        instances = bulk_mutate(mutation_datas, model=self.model)

        for instance, pre_data in zip(instances, previous_datas, strict=True):
            self.mutation_type.__after__(instance=instance, info=info, previous_data=pre_data)

        resolver = QueryTypeManyResolver(
            query_type=self.query_type,
            entrypoint=self.entrypoint,
            additional_filter=Q(pk__in=[instance.pk for instance in instances]),
        )
        return resolver.run_sync(root, info)

    async def run_async(self, root: Any, info: GQLInfo, **kwargs: Any) -> list[TModel]:
        # Fetch user eagerly so that its available e.g. for permission checks in synchronous parts of the code.
        await pre_evaluate_request_user(info)

        data: list[dict[str, Any]] = kwargs[undine_settings.MUTATION_INPUT_DATA_KEY]

        mutation_datas: list[MutationData] = await sync_to_async(get_mutation_data)(  # type: ignore[assignment]
            model=self.model,
            data=data,  # type: ignore[arg-type]
            mutation_type=self.mutation_type,
        )

        _pre_mutation_chain_many(
            root=root,
            info=info,
            mutation_datas=mutation_datas,
            mutation_type=self.mutation_type,
            permissions_func=self.entrypoint.permissions_func,
        )

        previous_datas = [mutation_data.previous_data for mutation_data in mutation_datas]

        instances: list[TModel] = await sync_to_async(bulk_mutate)(mutation_datas, model=self.model)  # type: ignore[arg-type,assignment]

        for instance, previous_data in zip(instances, previous_datas, strict=True):
            self.mutation_type.__after__(instance=instance, info=info, previous_data=previous_data)

        resolver = QueryTypeManyResolver(
            query_type=self.query_type,
            entrypoint=self.entrypoint,
            additional_filter=Q(pk__in=[instance.pk for instance in instances]),
        )
        return await resolver.run_async(root, info)


@dataclasses.dataclass(frozen=True, slots=True)
class BulkDeleteResolver(Generic[TModel]):
    """Resolves a bulk delete mutation for deleting a list of model instances."""

    mutation_type: type[MutationType]
    entrypoint: Entrypoint

    def __call__(self, root: Any, info: GQLInfo, **kwargs: Any) -> AwaitableOrValue[list[SimpleNamespace]]:
        if undine_settings.ASYNC:
            return self.run_async(root, info, **kwargs)
        return self.run_sync(root, info, **kwargs)

    @property
    def model(self) -> type[TModel]:
        return self.mutation_type.__model__  # type: ignore[return-value]

    def run_sync(self, root: Any, info: GQLInfo, **kwargs: Any) -> list[SimpleNamespace]:
        data: list[dict[str, Any]] = kwargs[undine_settings.MUTATION_INPUT_DATA_KEY]

        mutation_datas = get_mutation_data(
            model=self.model,
            data=data,
            mutation_type=self.mutation_type,
        )

        instances = [mutation_data.instance for mutation_data in mutation_datas]
        pks = [instance.pk for instance in instances]

        _pre_mutation_chain_many(
            root=root,
            info=info,
            mutation_datas=mutation_datas,
            mutation_type=self.mutation_type,
            permissions_func=self.entrypoint.permissions_func,
        )

        previous_datas = [mutation_data.previous_data for mutation_data in mutation_datas]

        with convert_integrity_errors(GraphQLModelConstraintViolationError):
            get_default_manager(self.model).filter(pk__in=pks).delete()

        for instance, previous_data in zip(instances, previous_datas, strict=True):
            self.mutation_type.__after__(instance=instance, info=info, previous_data=previous_data)

        return [SimpleNamespace(pk=pk) for pk in pks]

    async def run_async(self, root: Any, info: GQLInfo, **kwargs: Any) -> list[SimpleNamespace]:
        # Fetch user eagerly so that its available e.g. for permission checks in synchronous parts of the code.
        await pre_evaluate_request_user(info)

        data: list[dict[str, Any]] = kwargs[undine_settings.MUTATION_INPUT_DATA_KEY]

        mutation_datas: list[MutationData] = await sync_to_async(get_mutation_data)(  # type: ignore[assignment]
            model=self.model,
            data=data,  # type: ignore[arg-type]
            mutation_type=self.mutation_type,
        )

        instances = [mut_data.instance for mut_data in mutation_datas]
        pks = [instance.pk for instance in instances]

        _pre_mutation_chain_many(
            root=root,
            info=info,
            mutation_datas=mutation_datas,
            mutation_type=self.mutation_type,
            permissions_func=self.entrypoint.permissions_func,
        )

        previous_datas = [mutation_data.previous_data for mutation_data in mutation_datas]

        with convert_integrity_errors(GraphQLModelConstraintViolationError):
            await get_default_manager(self.model).filter(pk__in=pks).adelete()

        for instance, previous_data in zip(instances, previous_datas, strict=True):
            self.mutation_type.__after__(instance=instance, info=info, previous_data=previous_data)

        return [SimpleNamespace(pk=pk) for pk in pks]


# Custom


@dataclasses.dataclass(frozen=True, slots=True)
class CustomResolver:
    """Resolves a custom mutation a model instance."""

    mutation_type: type[MutationType]
    entrypoint: Entrypoint

    def __call__(self, root: Any, info: GQLInfo, **kwargs: Any) -> Any:
        if undine_settings.ASYNC and iscoroutinefunction(self.mutation_type.__mutate__):
            return self.run_async(root, info, **kwargs)
        return self.run_sync(root, info, **kwargs)

    @property
    def query_type(self) -> type[QueryType]:
        return self.mutation_type.__query_type__()

    @property
    def model(self) -> type[TModel]:
        return self.mutation_type.__model__  # type: ignore[return-value]

    def run_sync(self, root: Any, info: GQLInfo, **kwargs: Any) -> Any:
        data: dict[str, Any] = kwargs[undine_settings.MUTATION_INPUT_DATA_KEY]

        mutation_data = get_mutation_data(
            model=self.model,
            data=data,
            mutation_type=self.mutation_type,
        )

        _pre_mutation_chain(
            root=root,
            info=info,
            mutation_data=mutation_data,
            mutation_type=self.mutation_type,
            permissions_func=self.entrypoint.permissions_func,
        )

        previous_data = mutation_data.previous_data
        parent = mutation_data.instance
        input_data = mutation_data.plain_data

        with convert_integrity_errors(GraphQLModelConstraintViolationError):
            result = self.mutation_type.__mutate__(parent, info, input_data)

        if isinstance(result, self.model):
            self.mutation_type.__after__(instance=result, info=info, previous_data=previous_data)

            resolver = QueryTypeSingleResolver(query_type=self.query_type, entrypoint=self.entrypoint)
            return resolver.run_sync(root, info, pk=result.pk)

        return result

    async def run_async(self, root: Any, info: GQLInfo, **kwargs: Any) -> Any:
        # Fetch user eagerly so that its available e.g. for permission checks in synchronous parts of the code.
        await pre_evaluate_request_user(info)

        data: dict[str, Any] = kwargs[undine_settings.MUTATION_INPUT_DATA_KEY]

        mutation_data: MutationData = await sync_to_async(get_mutation_data)(  # type: ignore[assignment]
            model=self.model,
            data=data,  # type: ignore[arg-type]
            mutation_type=self.mutation_type,
        )

        _pre_mutation_chain(
            root=root,
            info=info,
            mutation_data=mutation_data,
            mutation_type=self.mutation_type,
            permissions_func=self.entrypoint.permissions_func,
        )

        previous_data = mutation_data.previous_data
        parent = mutation_data.instance
        input_data = mutation_data.plain_data

        with convert_integrity_errors(GraphQLModelConstraintViolationError):
            result = await self.mutation_type.__mutate__(parent, info, input_data)

        if isinstance(result, self.model):
            self.mutation_type.__after__(instance=result, info=info, previous_data=previous_data)

            resolver = QueryTypeSingleResolver(query_type=self.query_type, entrypoint=self.entrypoint)
            return await resolver.run_async(root, info, pk=result.pk)

        return result


# Helpers


def _pre_mutation_chain(
    root: Any,
    info: GQLInfo,
    mutation_data: MutationData,
    mutation_type: type[MutationType],
    permissions_func: InputPermFunc | EntrypointPermFunc | None = None,
) -> None:
    _add_hidden_inputs(
        mutation_data=mutation_data,
        mutation_type=mutation_type,
        info=info,
    )
    _check_permissions(
        parent=root,
        info=info,
        mutation_data=mutation_data,
        mutation_type=mutation_type,
        parent_permissions_func=permissions_func,
    )
    _validate(
        parent=root,
        info=info,
        mutation_data=mutation_data,
        mutation_type=mutation_type,
    )
    _remove_input_only_inputs(
        mutation_data=mutation_data,
        mutation_type=mutation_type,
        info=info,
    )


def _pre_mutation_chain_many(
    root: Any,
    info: GQLInfo,
    mutation_datas: list[MutationData],
    mutation_type: type[MutationType],
    permissions_func: InputPermFunc | EntrypointPermFunc | None = None,
) -> None:
    errors: list[GraphQLError] = []

    for i, mutation_data in enumerate(mutation_datas):
        try:
            with graphql_error_path(info, key=i) as sub_info:
                _pre_mutation_chain(
                    root=root,
                    info=sub_info,
                    mutation_data=mutation_data,
                    mutation_type=mutation_type,
                    permissions_func=permissions_func,
                )
        except GraphQLError as error:
            errors.append(error)

        except GraphQLErrorGroup as error_group:
            errors.extend(error_group.flatten())

    if errors:
        raise GraphQLErrorGroup(errors)


def _add_hidden_inputs(  # noqa: C901, PLR0912
    mutation_data: MutationData,
    mutation_type: type[MutationType],
    info: GQLInfo,
) -> None:
    instance = mutation_data.instance

    relation_info = parse_model_relation_info(model=mutation_type.__model__)

    for input_field in mutation_type.__input_map__.values():
        value = mutation_data.data.get(input_field.name, Undefined)

        if input_field.hidden:
            if isinstance(input_field.ref, FunctionType):
                value = input_field.ref(instance, info)
            elif input_field.default_value is not Undefined:
                value = input_field.default_value

        if value is Undefined:
            continue

        if not input_field.hidden and isinstance(input_field.ref, FunctionType):
            value = input_field.ref(instance, info, value)

        sub_mutation_type: type[MutationType] | None = None
        is_related_input = is_subclass(input_field.ref, MutationType)
        if is_related_input:
            sub_mutation_type = input_field.ref

        rel_info = relation_info.get(input_field.field_name)

        if rel_info is None:
            mutation_data.data[input_field.name] = value
            continue

        if rel_info.relation_type.is_single and not isinstance(value, MutationData):
            mutation_info = get_related_info(data=value, rel_info=rel_info)
            if mutation_info is not None:
                value = mutation_info.process(sub_mutation_type)

        elif rel_info.relation_type.is_many and not isinstance(value, MutationManyData):
            mut_info = [get_related_info(data=item, rel_info=rel_info) for item in value]
            mut_data = [item.process(sub_mutation_type) for item in mut_info if item is not None]
            value = MutationManyData(mutation_data=mut_data, mutation_type=sub_mutation_type)

        mutation_data.data[input_field.name] = value

        if not is_related_input:
            continue

        if isinstance(value, MutationManyData):
            for item in value:
                _add_hidden_inputs(item, input_field.ref, info)
            continue

        if isinstance(value, MutationData):
            _add_hidden_inputs(value, input_field.ref, info)


def _check_permissions(
    parent: Any,
    info: GQLInfo,
    mutation_data: MutationData,
    *,
    mutation_type: type[MutationType],
    parent_permissions_func: InputPermFunc | EntrypointPermFunc | None = None,
) -> None:
    input_data = mutation_data.plain_data
    instance = mutation_data.instance

    with graphql_error_path(info):
        if parent_permissions_func is not None:
            parent_permissions_func(parent, info, input_data)
        else:
            mutation_type.__permissions__(instance, info, input_data)

    errors: list[GraphQLError] = []

    for key, value in input_data.items():
        input_field = mutation_type.__input_map__[key]
        original_value = mutation_data.data[key]

        if value == input_field.default_value:
            continue

        try:
            with graphql_error_path(info, key=input_field.schema_name) as sub_info:
                _check_permissions_input(mutation_data.instance, sub_info, value, input_field, original_value)

        except GraphQLError as error:
            errors.append(error)

        except GraphQLErrorGroup as error_group:
            errors.extend(error_group.flatten())

    if errors:
        raise GraphQLErrorGroup(errors)


def _check_permissions_input(
    parent: Model,
    info: GQLInfo,
    value: Any,
    input_field: Input,
    original_value: Any,
) -> None:
    if is_subclass(input_field.ref, MutationType):
        if isinstance(original_value, MutationManyData):
            errors: list[GraphQLError] = []

            for index, item in enumerate(original_value):
                try:
                    with graphql_error_path(info, key=index) as list_info:
                        _check_permissions(
                            parent=parent,
                            info=list_info,
                            mutation_data=item,
                            mutation_type=input_field.ref,
                            parent_permissions_func=input_field.permissions_func,
                        )

                except GraphQLError as error:
                    errors.append(error)

                except GraphQLErrorGroup as error_group:
                    errors.extend(error_group.flatten())

            if errors:
                raise GraphQLErrorGroup(errors)

        elif isinstance(original_value, MutationData):
            _check_permissions(
                parent=parent,
                info=info,
                mutation_data=original_value,
                mutation_type=input_field.ref,
                parent_permissions_func=input_field.permissions_func,
            )

        # Custom mutations can use related mutation types for non-relational fields.
        elif input_field.mutation_type.__kind__ == MutationKind.custom:
            if input_field.permissions_func is not None:
                input_field.permissions_func(parent, info, value)

        else:
            msg = f"Unexpected type: {type(original_value)}"
            raise TypeError(msg)  # TODO: Custom Error

    elif input_field.permissions_func is not None:
        input_field.permissions_func(parent, info, value)


def _validate(
    parent: Any,
    info: GQLInfo,
    mutation_data: MutationData,
    *,
    mutation_type: type[MutationType],
    parent_validation_func: ValidatorFunc | None = None,
) -> None:
    errors: list[GraphQLError] = []

    input_data = mutation_data.plain_data
    instance = mutation_data.instance

    for key, value in input_data.items():
        input_field = mutation_type.__input_map__[key]
        original_value = mutation_data.data[key]

        if value == input_field.default_value:
            continue

        try:
            with graphql_error_path(info, key=input_field.schema_name) as sub_info:
                _validate_input(instance, sub_info, value, input_field, original_value)

        except GraphQLError as error:
            errors.append(error)

        except GraphQLErrorGroup as error_group:
            errors.extend(error_group.flatten())

    with graphql_error_path(info):
        if parent_validation_func is not None:
            parent_validation_func(parent, info, input_data)
        else:
            mutation_type.__validate__(instance, info, input_data)

    if errors:
        raise GraphQLErrorGroup(errors)


def _validate_input(
    parent: Model,
    info: GQLInfo,
    value: Any,
    input_field: Input,
    original_value: Any,
) -> None:
    if is_subclass(input_field.ref, MutationType):
        if isinstance(original_value, MutationManyData):
            errors: list[GraphQLError] = []

            for index, item in enumerate(original_value):
                try:
                    with graphql_error_path(info, key=index) as list_info:
                        _validate(
                            parent=parent,
                            info=list_info,
                            mutation_data=item,
                            mutation_type=input_field.ref,
                            parent_validation_func=input_field.validator_func,
                        )

                except GraphQLError as error:
                    errors.append(error)

                except GraphQLErrorGroup as error_group:
                    errors.extend(error_group.flatten())

            if errors:
                raise GraphQLErrorGroup(errors)

        elif isinstance(original_value, MutationData):
            _validate(
                parent=parent,
                info=info,
                mutation_data=original_value,
                mutation_type=input_field.ref,
                parent_validation_func=input_field.validator_func,
            )

        # Custom mutations can use related mutation types for non-relational fields.
        elif input_field.mutation_type.__kind__ == MutationKind.custom:
            if input_field.validator_func is not None:
                input_field.validator_func(parent, info, value)

        else:
            msg = f"Unexpected type: {type(value)}"
            raise TypeError(msg)  # TODO: Custom Error

    elif input_field.validator_func is not None:
        input_field.validator_func(parent, info, value)


def _remove_input_only_inputs(
    mutation_data: MutationData,
    mutation_type: type[MutationType],
    info: GQLInfo,
) -> None:
    for input_field in mutation_type.__input_map__.values():
        if input_field.input_only:
            mutation_data.data.pop(input_field.name, None)

        if input_field.name not in mutation_data.data:
            continue

        value = mutation_data.data[input_field.name]

        if not is_subclass(input_field.ref, MutationType):
            continue

        if isinstance(value, MutationManyData):
            for item in value:
                _remove_input_only_inputs(item, input_field.ref, info)
            continue

        if isinstance(value, MutationData):
            _remove_input_only_inputs(value, input_field.ref, info)



from __future__ import annotations

import dataclasses
from abc import ABC, abstractmethod
from collections.abc import AsyncGenerator, Awaitable, Callable, Generator
from contextlib import AsyncExitStack, ExitStack, asynccontextmanager, contextmanager
from functools import wraps
from typing import TYPE_CHECKING, Any, Self, TypeVar

if TYPE_CHECKING:
    from collections.abc import AsyncIterator
    from types import TracebackType

    from graphql import DocumentNode, ExecutionResult

    from undine.dataclasses import GraphQLHttpParams
    from undine.typing import DjangoRequestProtocol

__all__ = [
    "LifecycleHook",
    "LifecycleHookContext",
    "LifecycleHookManager",
    "delegate_to_subgenerator",
    "use_lifecycle_hooks_async",
    "use_lifecycle_hooks_sync",
]


@dataclasses.dataclass(slots=True, kw_only=True)
class LifecycleHookContext:
    """Context passed to a lifecycle hook."""

    source: str
    """Source GraphQL document string."""

    document: DocumentNode | None
    """Parsed GraphQL document AST. Available after parsing is complete."""

    variables: dict[str, Any]
    """Variables passed to the GraphQL operation."""

    operation_name: str | None
    """Name of the GraphQL operation."""

    extensions: dict[str, Any]
    """GraphQL operation extensions received from the client."""

    request: DjangoRequestProtocol
    """Django request during which the GraphQL request is being executed."""

    result: ExecutionResult | Awaitable[ExecutionResult | AsyncIterator[ExecutionResult]] | None
    """Execution result of the GraphQL operation. Adding a result here will cause an early exit."""

    @classmethod
    def from_graphql_params(cls, params: GraphQLHttpParams, request: DjangoRequestProtocol) -> Self:
        return cls(
            source=params.document,
            document=None,
            variables=params.variables,
            operation_name=params.operation_name,
            extensions=params.extensions,
            request=request,
            result=None,
        )


@dataclasses.dataclass(frozen=True, slots=True, kw_only=True)
class LifecycleHook(ABC):
    """Base class for lifecycle hooks."""

    context: LifecycleHookContext

    @contextmanager
    def use_sync(self) -> Generator[None, None, None]:
        yield from self.run()

    @asynccontextmanager
    async def use_async(self) -> AsyncGenerator[None, None]:
        gen = self.run_async()
        async with delegate_to_subgenerator(gen):
            async for _ in gen:
                yield

    @abstractmethod
    def run(self) -> Generator[None, None, None]:
        """
        Override this method to define how the hook should be executed.
        Anything before the yield statement will be executed before the hooking point.
        Anything after the yield statement will be executed after the hooking point.
        """
        yield

    async def run_async(self) -> AsyncGenerator[None, None]:
        """
        Override this method to define how the hook should be executed in an async context.
        Uses the `run` method by default.
        """
        with delegate_to_subgenerator(self.run()) as gen:
            for _ in gen:
                yield


TLifecycleHook = TypeVar("TLifecycleHook", bound=LifecycleHook)


class LifecycleHookManager(ExitStack, AsyncExitStack):
    """Allows executing multiple lifecycle hooks at once."""

    def __init__(self, *, hooks: list[type[TLifecycleHook]], context: LifecycleHookContext) -> None:
        self.hooks: list[TLifecycleHook] = [hook(context=context) for hook in hooks]
        super().__init__()

    def __enter__(self) -> Self:
        for hook in self.hooks:
            self.enter_context(hook.use_sync())
        return super().__enter__()

    async def __aenter__(self) -> Self:
        for hook in self.hooks:
            await self.enter_async_context(hook.use_async())
        return await super().__aenter__()


R = TypeVar("R")
HookableSync = Callable[[LifecycleHookContext], R]
HookableAsync = Callable[[LifecycleHookContext], Awaitable[R]]


def use_lifecycle_hooks_sync(hooks: list[type[TLifecycleHook]]) -> Callable[[HookableSync[R]], HookableSync[R]]:
    """Run given function using the given lifecycle hooks."""

    def decorator(func: HookableSync[R]) -> HookableSync[R]:
        @wraps(func)
        def wrapper(context: LifecycleHookContext) -> R:
            with LifecycleHookManager(hooks=hooks, context=context):
                return func(context)

        return wrapper

    return decorator


def use_lifecycle_hooks_async(hooks: list[type[TLifecycleHook]]) -> Callable[[HookableAsync[R]], HookableAsync[R]]:
    """Run given function using the given lifecycle hooks."""

    def decorator(func: HookableAsync[R]) -> HookableAsync[R]:
        @wraps(func)
        async def wrapper(context: LifecycleHookContext) -> R:  # type: ignore[return]
            async with LifecycleHookManager(hooks=hooks, context=context):
                return await func(context)

        return wrapper

    return decorator


class delegate_to_subgenerator:  # noqa: N801
    """
    Allows delegating how a generator exists to a subgenerator.

    >>> def subgenerator():
    ...     for _ in range(2):
    ...         yield
    >>>
    >>> def generator():
    >>>     with delegate_to_subgenerator(subgenerator()) as sub:
    ...         for _ in sub:
    ...             yield
    >>>
    >>> for item in generator():
    ...     pass

    If the generator exists normally, the subgenerator will be closed.
    If the generator exists with an exception, the error is propagated to the subgenerator
    so that it may handle the error.
    """

    def __init__(self, gen: Generator[None, None, None] | AsyncGenerator[None, None]) -> None:
        """
        Allows delegating how a generator exists to a subgenerator.

        :param gen: The generator to delegate to. If generator is an async generator,
                    must use `async with` syntax to delegate. For regular generators,
                    plain `with` syntax must be used.
        """
        self.gen = gen

    def __enter__(self) -> Generator[None, None, None]:
        if not isinstance(self.gen, Generator):
            msg = "Given object is not a Generator"
            raise TypeError(msg)

        return self.gen

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool:
        if not isinstance(self.gen, Generator):  # type: ignore[unreachable]
            msg = "Given object is not a Generator"
            raise TypeError(msg)

        # If no exception was raised, close the generator.
        if exc_type is None:
            self.gen.close()
            return False

        # Otherwise, allow the subgenerator to handle the exception.
        # This has mostly been copied from `contextlib._GeneratorContextManager.__exit__`.
        if exc_value is None:
            exc_value = exc_type()

        try:
            self.gen.throw(exc_value)

        except StopIteration as error:
            return error is not exc_value

        except RuntimeError as error:
            if error is exc_value:
                error.__traceback__ = traceback
                return False
            if isinstance(exc_value, StopIteration) and error.__cause__ is exc_value:
                exc_value.__traceback__ = traceback
                return False
            raise

        except BaseException as error:
            if error is not exc_value:
                raise
            error.__traceback__ = traceback
            return False

        try:
            msg = "generator didn't stop after throw()"
            raise RuntimeError(msg)
        finally:
            self.gen.close()

    async def __aenter__(self) -> AsyncGenerator[None, None]:
        if not isinstance(self.gen, AsyncGenerator):
            msg = "Given object is not an AsyncGenerator"
            raise TypeError(msg)

        return self.gen

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool:
        if not isinstance(self.gen, AsyncGenerator):
            msg = "Given object is not an AsyncGenerator"
            raise TypeError(msg)

        # If no exception was raised, close the generator.
        if exc_type is None:
            await self.gen.aclose()
            return False

        # Otherwise, allow the subgenerator to handle the exception.
        # This has mostly been copied from `contextlib._AsyncGeneratorContextManager.__aexit__`.
        if exc_value is None:
            exc_value = exc_type()

        try:
            await self.gen.athrow(exc_value)

        except StopAsyncIteration as error:
            return error is not exc_value

        except RuntimeError as error:
            if error is exc_value:
                error.__traceback__ = traceback
                return False
            if isinstance(exc_value, (StopIteration, StopAsyncIteration)) and error.__cause__ is exc_value:
                exc_value.__traceback__ = traceback
                return False
            raise

        except BaseException as error:
            if error is not exc_value:
                raise
            error.__traceback__ = traceback
            return False

        try:
            msg = "generator didn't stop after athrow()"
            raise RuntimeError(msg)
        finally:
            await self.gen.aclose()
